'''
**Module to generate multibunch distributions**

:Authors: **Danilo Quartullo**, **Alexandre Lasheen**, **Theodoros Argyropoulos**
'''

from __future__ import absolute_import, division, print_function

import copy
import gc
from builtins import range

import matplotlib.pyplot as plt
import numpy as np
import scipy
from packaging.version import Version

if Version(scipy.__version__) >= Version("1.14"):
    from scipy.integrate import cumulative_trapezoid as cumtrapz
else:
    from scipy.integrate import cumtrapz


from ..beam.beam import Beam
from ..beam.distributions import (X0_from_bunch_length, distribution_function,
                                  matched_from_distribution_function,
                                  matched_from_line_density, populate_bunch,
                                  potential_well_cut)
from ..utils import bmath as bm


def matched_from_distribution_density_multibunch(beam, Ring, FullRingAndRF, distribution_options_list,
                                                 n_bunches, bunch_spacing_buckets,
                                                 intensity_list=None,
                                                 minimum_n_macroparticles=None,
                                                 main_harmonic_option='lowest_freq',
                                                 TotalInducedVoltage=None,
                                                 n_iterations_input=1,
                                                 plot_option=False, seed=None):
    '''
    *Function to generate a multi-bunch beam using the matched_from_distribution_density
    function for each bunch. The extra parameters to include are the number of
    bunches and the spacing between two bunches (assumed constant presently).
    Moreover, the distribution_options_list corresponds to the distribution_options
    of the matched_from_distribution_density function. It can be inputed as
    a dictionary just like the matched_from_distribution_density function (assuming
    the same parameters for all bunches), or as a list of length n_bunches
    to have different parameters for each bunch.*
    '''

    if intensity_list is None:
        intensity_per_bunch = beam.intensity / n_bunches * np.ones(n_bunches)
        n_macroparticles_per_bunch = beam.n_macroparticles / n_bunches * np.ones(n_bunches)
    else:
        intensity_per_bunch = np.array(intensity_list)
        if minimum_n_macroparticles is None:
            n_macroparticles_per_bunch = np.round(beam.n_macroparticles / beam.intensity * intensity_per_bunch)
        else:
            n_macroparticles_per_bunch = np.round(minimum_n_macroparticles / np.min(intensity_per_bunch) * intensity_per_bunch)

    if np.sum(intensity_per_bunch) != beam.intensity:
        print('WARNING !! The total intensity per bunch does not match the total intensity of the beam, the beam.intensity will be overwritten')
        beam.intensity = np.sum(intensity_per_bunch)

    if np.sum(n_macroparticles_per_bunch) != beam.n_macroparticles:
        print('WARNING !! The number of macroparticles per bunch does not match the total number of the beam, the beam.n_macroparticles will be overwritten')
        beam.n_macroparticles = int(np.sum(n_macroparticles_per_bunch))

    voltages = np.array([])
    harmonics = np.array([])

    for RingAndRFSectionElement in FullRingAndRF.RingAndRFSection_list:
        for rf_system in range(RingAndRFSectionElement.rf_params.n_rf):
            voltages = np.append(voltages, RingAndRFSectionElement.rf_params.voltage[rf_system, 0])
            harmonics = np.append(harmonics, RingAndRFSectionElement.rf_params.harmonic[rf_system, 0])

    if main_harmonic_option == 'lowest_freq':
        main_harmonic = np.min(harmonics)
    elif main_harmonic_option == 'highest_voltage':
        main_harmonic = np.min(harmonics[voltages == np.max(voltages)])
    elif isinstance(main_harmonic_option, int) or isinstance(main_harmonic_option, float):
        if harmonics[harmonics == main_harmonic_option].size == 0:
            # GenerationError
            raise RuntimeError('The desired harmonic to compute the potential well does not match the RF parameters...')
        main_harmonic = np.min(harmonics[harmonics == main_harmonic_option])

    bucket_size_tau = 2 * np.pi / (main_harmonic * Ring.omega_rev[0])

    beamIteration = Beam(Ring, 1, 0.)

    extraVoltageDict = None

    if TotalInducedVoltage is not None:
        bucket_tolerance = 0.40
        TotalInducedVoltageIteration = copy.deepcopy(TotalInducedVoltage)
        TotalInducedVoltageIteration.profile.Beam = beamIteration

    for indexBunch in range(0, n_bunches):

        print('Generating bunch no %d' % (indexBunch + 1))

        bunch = Beam(Ring, int(n_macroparticles_per_bunch[indexBunch]), intensity_per_bunch[indexBunch])

        if isinstance(distribution_options_list, list):
            distribution_options = distribution_options_list[indexBunch]
        elif isinstance(distribution_options_list, dict):
            distribution_options = distribution_options_list
        else:
            # DistributionError
            raise RuntimeError('The input distribution_options_list option of the matched_from_distribution_density_multibunch \
            function should either be a dictionary as requested by the matched_from_distribution_density \
            function, or a list of dictionaries containing n_bunches elements')

        if 'type' in distribution_options:
            distribution_type = distribution_options['type']
        else:
            distribution_type = None

        if 'exponent' in distribution_options:
            distribution_exponent = distribution_options['exponent']
        else:
            distribution_exponent = None

        if 'emittance' in distribution_options:
            emittance = distribution_options['emittance']
        else:
            emittance = None

        if 'bunch_length' in distribution_options:
            bunch_length = distribution_options['bunch_length']
        else:
            bunch_length = None

        if 'bunch_length_fit' in distribution_options:
            bunch_length_fit = distribution_options['bunch_length_fit']
        else:
            bunch_length_fit = None

        if 'density_variable' in distribution_options:
            distribution_variable = distribution_options['density_variable']
        else:
            distribution_variable = None

        if distribution_options['type'] == 'user_input':
            distribution_function_input = distribution_options['function']
        else:
            distribution_function_input = None

        if distribution_options['type'] == 'user_input_table':
            distribution_user_table = {
                'user_table_action': distribution_options['user_table_action'],
                'user_table_density': distribution_options['user_table_density']}
        else:
            distribution_user_table = None

        matched_from_distribution_function(bunch, FullRingAndRF,
                                           distribution_function_input=distribution_function_input,
                                           distribution_user_table=distribution_user_table,
                                           main_harmonic_option=main_harmonic_option,
                                           TotalInducedVoltage=TotalInducedVoltage,
                                           n_iterations=n_iterations_input,
                                           extraVoltageDict=extraVoltageDict,
                                           distribution_exponent=distribution_exponent,
                                           distribution_type=distribution_type,
                                           emittance=emittance, bunch_length=bunch_length,
                                           bunch_length_fit=bunch_length_fit,
                                           distribution_variable=distribution_variable, seed=seed)

        if indexBunch == 0:
            beamIteration.dt = bunch.dt
            beamIteration.dE = bunch.dE
        else:
            beamIteration.dt = np.append(beamIteration.dt, bunch.dt +
                                         (indexBunch * bunch_spacing_buckets * bucket_size_tau))
            beamIteration.dE = np.append(beamIteration.dE, bunch.dE)

        beamIteration.n_macroparticles = int(np.sum(n_macroparticles_per_bunch[:indexBunch + 1]))
        beamIteration.intensity = np.sum(intensity_per_bunch[:indexBunch + 1])
        beamIteration.ratio = beamIteration.intensity / beamIteration.n_macroparticles

        if TotalInducedVoltage is not None:
            TotalInducedVoltageIteration.profile.track()
            TotalInducedVoltageIteration.induced_voltage_sum()

            left_edge = ((indexBunch + 1) * bunch_spacing_buckets *
                         bucket_size_tau - bucket_tolerance * bucket_size_tau)
            right_edge = (((indexBunch + 1) * bunch_spacing_buckets + 1) *
                          bucket_size_tau + bucket_tolerance * bucket_size_tau)

            bin_centers = TotalInducedVoltageIteration.profile.bin_centers

            tau_induced_voltage_next_bunch = bin_centers[
                (bin_centers > left_edge) * (bin_centers < right_edge)]
            induced_voltage_next_bunch = \
                TotalInducedVoltageIteration.induced_voltage[
                    (bin_centers > left_edge) * (bin_centers < right_edge)]

            time_induced_voltage_next_bunch = (tau_induced_voltage_next_bunch -
                                               (indexBunch + 1) * bunch_spacing_buckets * bucket_size_tau)

            extraVoltageDict = {'time_array': time_induced_voltage_next_bunch,
                                'voltage_array': induced_voltage_next_bunch}

        if plot_option:
            plt.figure('Bunch train + induced voltage')
            plt.clf()
            plt.plot(TotalInducedVoltageIteration.profile.bin_centers,
                     TotalInducedVoltageIteration.profile.n_macroparticles /
                     (1. * np.max(TotalInducedVoltageIteration.profile.n_macroparticles)) *
                     np.max(TotalInducedVoltageIteration.induced_voltage))
            plt.plot(TotalInducedVoltageIteration.profile.bin_centers,
                     TotalInducedVoltageIteration.induced_voltage)
            plt.show()

    beam.dt = beamIteration.dt.astype(dtype=bm.precision.real_t, order='C', copy=False)
    beam.dE = beamIteration.dE.astype(dtype=bm.precision.real_t, order='C', copy=False)
    gc.collect()


def matched_from_line_density_multibunch(beam, Ring,
                                         FullRingAndRF, line_density_options_list, n_bunches,
                                         bunch_spacing_buckets, intensity_list=None,
                                         minimum_n_macroparticles=None,
                                         main_harmonic_option='lowest_freq',
                                         TotalInducedVoltage=None, half_option='first',
                                         plot_option=False, seed=None):
    '''
    *Function to generate a multi-bunch beam using the matched_from_distribution_density
    function for each bunch. The extra parameters to include are the number of
    bunches and the spacing between two bunches (assumed constant presently).
    Moreover, the line_density_options_list corresponds to the distribution_options
    of the matched_from_line_density function. It can be inputed as
    a dictionary just like the matched_from_line_density function (assuming
    the same parameters for all bunches), or as a list of length n_bunches
    to have different parameters for each bunch.*
    '''

    if intensity_list is None:
        intensity_per_bunch = beam.intensity / n_bunches * np.ones(n_bunches)
        n_macroparticles_per_bunch = beam.n_macroparticles / n_bunches * np.ones(n_bunches)
    else:
        intensity_per_bunch = np.array(intensity_list)
        if minimum_n_macroparticles is None:
            n_macroparticles_per_bunch = np.round(beam.n_macroparticles / beam.intensity * intensity_per_bunch)
        else:
            n_macroparticles_per_bunch = np.round(minimum_n_macroparticles / np.min(intensity_per_bunch) * intensity_per_bunch)

    if np.sum(intensity_per_bunch) != beam.intensity:
        print('WARNING !! The total intensity per bunch does not match the total intensity of the beam, the beam.intensity will be overwritten')
        beam.intensity = np.sum(intensity_per_bunch)

    if np.sum(n_macroparticles_per_bunch) != beam.n_macroparticles:
        print('WARNING !! The number of macroparticles per bunch does not match the total number of the beam, the beam.n_macroparticles will be overwritten')
        beam.n_macroparticles = int(np.sum(n_macroparticles_per_bunch))

    voltages = np.array([])
    harmonics = np.array([])

    for RingAndRFSectionElement in FullRingAndRF.RingAndRFSection_list:
        for rf_system in range(RingAndRFSectionElement.rf_params.n_rf):
            voltages = np.append(voltages, RingAndRFSectionElement.rf_params.voltage[rf_system, 0])
            harmonics = np.append(harmonics, RingAndRFSectionElement.rf_params.harmonic[rf_system, 0])

    if main_harmonic_option == 'lowest_freq':
        main_harmonic = np.min(harmonics)
    elif main_harmonic_option == 'highest_voltage':
        main_harmonic = np.min(harmonics[voltages == np.max(voltages)])
    elif isinstance(main_harmonic_option, int) or isinstance(main_harmonic_option, float):
        if harmonics[harmonics == main_harmonic_option].size == 0:
            # GenerationError
            raise RuntimeError('The desired harmonic to compute the potential well does not match the RF parameters...')
        main_harmonic = np.min(harmonics[harmonics == main_harmonic_option])

    bucket_size_tau = 2 * np.pi / (main_harmonic * Ring.omega_rev[0])

    beamIteration = Beam(Ring, 1, 0.)

    extraVoltageDict = None

    if TotalInducedVoltage is not None:
        TotalInducedVoltageIteration = copy.deepcopy(TotalInducedVoltage)
        TotalInducedVoltageIteration.profile.Beam = beamIteration

    for indexBunch in range(0, n_bunches):

        print('Generating bunch no %d' % (indexBunch + 1))

        bunch = Beam(Ring,
                     int(n_macroparticles_per_bunch[indexBunch]),
                     intensity_per_bunch[indexBunch])

        if isinstance(line_density_options_list, list):
            line_density_options = line_density_options_list[indexBunch]
        elif isinstance(line_density_options_list, dict):
            line_density_options = line_density_options_list
        else:
            # GenerationError
            raise RuntimeError('The input line_density_options_list option ' +
                               'of the matched_from_line_density_multibunch ' +
                               'function should either be a dictionary as ' +
                               'requested by the matched_from_line_density ' +
                               'function, or a list of dictionaries ' +
                               'containing n_bunches elements')

        if 'bunch_length' in line_density_options:
            bunch_length = line_density_options['bunch_length']
        else:
            bunch_length = None

        if 'type' in line_density_options:
            line_density_type = line_density_options['type']
        else:
            line_density_type = None

        if 'exponent' in line_density_options:
            line_density_exponent = line_density_options['exponent']
        else:
            line_density_exponent = None

        if line_density_options['type'] == 'user_input':
            line_density_input = {
                'time_line_den': line_density_options['time_line_den'],
                'line_density': line_density_options['line_density']}
        else:
            line_density_input = None

        matched_from_line_density(bunch, FullRingAndRF,
                                  line_density_input=line_density_input,
                                  main_harmonic_option=main_harmonic_option,
                                  TotalInducedVoltage=TotalInducedVoltage,
                                  plot=plot_option, half_option=half_option,
                                  extraVoltageDict=extraVoltageDict,
                                  bunch_length=bunch_length,
                                  line_density_type=line_density_type,
                                  line_density_exponent=line_density_exponent,
                                  seed=seed)

        if indexBunch == 0:
            beamIteration.dt = bunch.dt
            beamIteration.dE = bunch.dE
        else:
            beamIteration.dt = np.append(beamIteration.dt, bunch.dt +
                                         (indexBunch * bunch_spacing_buckets * bucket_size_tau))
            beamIteration.dE = np.append(beamIteration.dE, bunch.dE)

        beamIteration.n_macroparticles = int(np.sum(n_macroparticles_per_bunch[:indexBunch + 1]))
        beamIteration.intensity = np.sum(intensity_per_bunch[:indexBunch + 1])
        beamIteration.ratio = beamIteration.intensity / beamIteration.n_macroparticles

        if TotalInducedVoltage is not None:
            TotalInducedVoltageIteration.profile.track()
            TotalInducedVoltageIteration.induced_voltage_sum()

            bucket_tolerance = 0.40

            left_edge = (indexBunch + 1) * bunch_spacing_buckets * bucket_size_tau - bucket_tolerance * bucket_size_tau
            right_edge = ((indexBunch + 1) * bunch_spacing_buckets + 1) * bucket_size_tau + bucket_tolerance * bucket_size_tau

            tau_induced_voltage_next_bunch = TotalInducedVoltageIteration.profile.bin_centers[(TotalInducedVoltageIteration.profile.bin_centers > left_edge) * (TotalInducedVoltageIteration.profile.bin_centers < right_edge)]
            induced_voltage_next_bunch = TotalInducedVoltageIteration.induced_voltage[(TotalInducedVoltageIteration.profile.bin_centers > left_edge) * (TotalInducedVoltageIteration.profile.bin_centers < right_edge)]

            time_induced_voltage_next_bunch = (tau_induced_voltage_next_bunch - (indexBunch + 1) * bunch_spacing_buckets * bucket_size_tau)

            extraVoltageDict = {'time_array': time_induced_voltage_next_bunch, 'voltage_array': induced_voltage_next_bunch}

    if plot_option:
        plt.figure('Bunch train + induced voltage')
        plt.clf()
        plt.plot(TotalInducedVoltageIteration.profile.bin_centers, TotalInducedVoltageIteration.profile.n_macroparticles / (1. * np.max(TotalInducedVoltageIteration.profile.n_macroparticles)) * np.max(TotalInducedVoltageIteration.induced_voltage))
        plt.plot(TotalInducedVoltageIteration.profile.bin_centers, TotalInducedVoltageIteration.induced_voltage)
        plt.show()

    beam.dt = beamIteration.dt.astype(dtype=bm.precision.real_t, order='C', copy=False)
    beam.dE = beamIteration.dE.astype(dtype=bm.precision.real_t, order='C', copy=False)
    gc.collect()


def match_beam_from_distribution(beam, FullRingAndRF, GeneralParameters,
                                 distribution_options, n_bunches,
                                 bunch_spacing_buckets,
                                 main_harmonic_option='lowest_freq',
                                 TotalInducedVoltage=None, n_iterations=1,
                                 n_points_potential=1e4,
                                 dt_margin_percent=0.40, seed=None,):
    '''
    *This function generates n equaly spaced bunches for a stationary
    distribution and try to match them with intensity effects.*

    *The corresponding distributions are specified by their exponent:*

    .. math::
        g_0(J) \\sim (1-J/J_0)^{\\text{exponent}}}

    *Knowing the distribution, to generate the phase space:
    - Compute the potential U
    - The value of H can be computed thanks to U
    - The action J can be integrated over the whole phase space
    - 2piJ = emittance, this restrict the value of J0 (or H0)
    - with g0(H) we can randomize the macroparticles*
    '''
# ------------------------------------------------------------------------
# USEFUL VARIABLES
# ------------------------------------------------------------------------
    # Slicing necessary only with intensity effects
    if TotalInducedVoltage is not None:
        profile = TotalInducedVoltage.profile

    # Ring informations, Trev, energy, RF parameters ...
    rf_params = FullRingAndRF.RingAndRFSection_list[0].rf_params
    t_rev = rf_params.t_rev[0]
    n_rf = rf_params.n_rf
    beta = rf_params.beta[0]
    E = rf_params.energy[0]
    charge = rf_params.charge
#    acceleration_kick = FullRingAndRF.RingAndRFSection_list[0].acceleration_kick[0]

    # Minimum omega_rf is used to compute the size of the bucket
    omega_rf = []
    for i in range(n_rf):
        omega_rf += [rf_params.omega_rf[i][0]]
    omega_rf = np.array(omega_rf)

    eta_0 = rf_params.eta_0[0]

    # Coefficient of Kin and Pot part of the hamiltonian
    normalization_DeltaE = np.abs(eta_0) / (2. * beta**2 * E)
    normalization_potential = np.sign(eta_0) * charge / t_rev

    intensity_per_bunch = beam.intensity / n_bunches
    n_macro_per_bunch = int(beam.n_macroparticles / n_bunches)
    bucket_size_tau = 2 * np.pi / (np.min(omega_rf))

# ------------------------------------------------------------------------
# GENERATES N BUNCHES WITHOUT INTENSITY EFFECTS
# ------------------------------------------------------------------------

    FullRingAndRF.potential_well_generation(n_points=n_points_potential,
                                            dt_margin_percent=dt_margin_percent,
                                            main_harmonic_option=main_harmonic_option)

    # Restrict the potential well inside the separatrix and put min on 0
    potential_well_coordinates, potential_well = potential_well_cut(
        FullRingAndRF.potential_well_coordinates,
        FullRingAndRF.potential_well)
    potential_well = potential_well - np.min(potential_well)

    # Temporary beam, everything is done in the first bucket and then
    # shifted to plug into the real beam.
    temporary_beam = Beam(GeneralParameters, n_macro_per_bunch, intensity_per_bunch)

    # Bunches placed in all the buckets without intensity effects
    # Loop the match function to have "different" bunches in each bucket
    matched_bunch_list = []
    for indexBunch in range(n_bunches):
        (time_grid, deltaE_grid, distribution, time_resolution,
            energy_resolution, single_profile) = match_a_bunch(
                normalization_DeltaE, temporary_beam,
                potential_well_coordinates,
                potential_well, seed, distribution_options,
                full_ring_and_RF=FullRingAndRF)
        matched_bunch_list.append(
            (time_grid, deltaE_grid, distribution, time_resolution,
             energy_resolution, single_profile))

    print(str(n_bunches) + ' stationary bunches without intensity generated')
# ------------------------------------------------------------------------
# REMATCH THE BUNCHES WITH INTENSITY EFFECTS
# ------------------------------------------------------------------------
    if TotalInducedVoltage is not None:
        print('Applying intensity effects ...')
        previous_well = potential_well
        for it in range(n_iterations):
            conv = 0.
            # Compute the induced voltage/potential for all the beam
            profile.n_macroparticles[:] = 0
            for indexBunch in range(n_bunches):
                profile.n_macroparticles += np.interp(
                    profile.bin_centers,
                    potential_well_coordinates +
                    indexBunch * bunch_spacing_buckets * bucket_size_tau,
                    matched_bunch_list[indexBunch][5],
                    left=0, right=0)
            profile.n_macroparticles[:] *= 1 / (np.sum(profile.n_macroparticles)) * beam.n_macroparticles

            TotalInducedVoltage.induced_voltage_sum()

            induced_voltage_coordinates = TotalInducedVoltage.time_array
            induced_voltage = TotalInducedVoltage.induced_voltage
            induced_potential = - normalization_potential * cumtrapz(
                induced_voltage,
                dx=induced_voltage_coordinates[1] -
                induced_voltage_coordinates[0],
                initial=0)

            for indexBunch in range(n_bunches):
                # Extract the induced potential for the specific bucket
                induced_potential_bunch = np.interp(potential_well_coordinates
                                                    + indexBunch * bunch_spacing_buckets * bucket_size_tau,
                                                    induced_voltage_coordinates, induced_potential)

                distorted_pot_well = potential_well + induced_potential_bunch
                distorted_pot_well -= np.min(distorted_pot_well)

                # Recompute the phase space distribution for the new
                # perturbed potential (containing induced_potential_bunch)
                matched_bunch_list[indexBunch] = match_a_bunch(
                    normalization_DeltaE, temporary_beam,
                    potential_well_coordinates,
                    distorted_pot_well, seed,
                    distribution_options,
                    full_ring_and_RF=FullRingAndRF)

            conv = np.sqrt(np.sum((previous_well - distorted_pot_well)**2.)) / len(distorted_pot_well)
            previous_well = distorted_pot_well

            print('iteration ' + str(it + 1) + ', convergence parameter = ' + str(conv))

            profile.n_macroparticles[:] = 0
            for indexBunch in range(n_bunches):
                profile.n_macroparticles += np.interp(
                    profile.bin_centers,
                    potential_well_coordinates +
                    indexBunch * bunch_spacing_buckets * bucket_size_tau,
                    matched_bunch_list[indexBunch][5],
                    left=0, right=0)
            profile.n_macroparticles[:] *= 1 / (np.sum(profile.n_macroparticles)) * beam.n_macroparticles

            TotalInducedVoltage.induced_voltage_sum()

    for indexBunch in range(n_bunches):

        (time_grid, deltaE_grid, distribution, time_resolution,
         energy_resolution, single_profile) = matched_bunch_list[indexBunch]
        populate_bunch(temporary_beam, time_grid, deltaE_grid, distribution,
                       time_resolution, energy_resolution, seed)

        length_dt = len(temporary_beam.dt)
        length_dE = len(temporary_beam.dE)

        beam.dt[indexBunch * length_dt:(indexBunch + 1) * length_dt] = np.array(
            temporary_beam.dt) + (indexBunch * bunch_spacing_buckets * bucket_size_tau)
        beam.dE[indexBunch * length_dE:(indexBunch + 1) * length_dE] = np.array(
            temporary_beam.dE)

    beam.dt = beam.dt.astype(dtype=bm.precision.real_t, order='C', copy=False)
    beam.dE = beam.dE.astype(dtype=bm.precision.real_t, order='C', copy=False)
    gc.collect()


def match_beam_from_distribution_multibatch(beam, FullRingAndRF, GeneralParameters,
                                            distribution_options, n_bunches,
                                            bunch_spacing_buckets, n_batch,
                                            batch_spacing_buckets,
                                            main_harmonic_option='lowest_freq',
                                            TotalInducedVoltage=None, n_iterations=1,
                                            n_points_potential=1e4,
                                            dt_margin_percent=0.40, seed=None):
    '''
    *This function generates n equaly spaced bunches for a stationary
    distribution and try to match them with intensity effects.*

    *Then it copies the batch n_batch times with spacing batch_spacing_buckets*

    *The corresponding distributions are specified by their exponent:*

    .. math::
        g_0(J) \\sim (1-J/J_0)^{\\text{exponent}}}

    *Knowing the distribution, to generate the phase space:
    - Compute the potential U
    - The value of H can be computed thanks to U
    - The action J can be integrated over the whole phase space
    - 2piJ = emittance, this restrict the value of J0 (or H0)
    - with g0(H) we can randomize the macroparticles*
    '''
# ------------------------------------------------------------------------
# USEFUL VARIABLES
# ------------------------------------------------------------------------
    # Ring informations, Trev, energy, RF parameters ...
    rf_params = FullRingAndRF.RingAndRFSection_list[0].rf_params
    n_rf = rf_params.n_rf
    # Slicing necessary only with intensity effects
    if TotalInducedVoltage is not None:
        profile = TotalInducedVoltage.profile

        t_rev = rf_params.t_rev[0]
        beta = rf_params.beta[0]
        E = rf_params.energy[0]
        charge = rf_params.charge
        eta_0 = rf_params.eta_0[0]

        normalization_DeltaE = np.abs(eta_0) / (2. * beta**2 * E)
        normalization_potential = np.sign(eta_0) * charge / t_rev

    # Ring informations, Trev, energy, RF parameters ...
#    beta = rf_params.beta[0]
#    E = rf_params.energy[0]
#    charge = rf_params.charge
#    acceleration_kick = FullRingAndRF.RingAndRFSection_list[0].acceleration_kick[0]

    # Minimum omega_rf is used to compute the size of the bucket
    omega_rf = []
    for i in range(n_rf):
        omega_rf += [rf_params.omega_rf[i][0]]
    omega_rf = np.array(omega_rf)

#    eta_0 = rf_params.eta_0[0]

#    # Coefficient of Kin and Pot part of the hamiltonian
#    normalization_DeltaE = np.abs(eta_0) / (2.*beta**2*E)
#    normalization_potential = np.sign(eta_0)*charge/t_rev

    intensity_per_bunch = beam.intensity / n_bunches / n_batch
    n_macro_per_bunch = int(beam.n_macroparticles / n_bunches / n_batch)
    bucket_size_tau = 2 * np.pi / (np.min(omega_rf))

    temporary_batch = Beam(GeneralParameters, int(n_macro_per_bunch * n_bunches), (intensity_per_bunch * n_bunches))

#    print(temporary_batch.dt)
    match_beam_from_distribution(temporary_batch, FullRingAndRF, GeneralParameters,
                                 distribution_options, n_bunches, bunch_spacing_buckets,
                                 TotalInducedVoltage=None, n_iterations=n_iterations,
                                 n_points_potential=n_points_potential)

#    matched_from_distribution_density_multibunch(temporary_batch, GeneralParameters, FullRingAndRF, distribution_options,
#                                          n_bunches, bunch_spacing_buckets,
#                                          TotalInducedVoltage = TotalInducedVoltage,
#                                          n_iterations_input = n_iterations)
    length_dt = len(temporary_batch.dt)
    print(length_dt)
    for index_batch in range(n_batch):
        beam.dt[index_batch * length_dt:(index_batch + 1) * length_dt] = temporary_batch.dt + index_batch * (n_bunches - 1) * bunch_spacing_buckets * bucket_size_tau + (index_batch) * batch_spacing_buckets * bucket_size_tau
        beam.dE[index_batch * length_dt:(index_batch + 1) * length_dt] = temporary_batch.dE

    plt.figure('copymultibatch')
    plt.plot(beam.dt[::100], beam.dE[::100], 'b.')
    plt.figure('temporarybatch')
    plt.plot(temporary_batch.dt[::100], temporary_batch.dE[::100], 'b.')
    plt.figure('profile before induced voltage')
    profile.track()
    plt.plot(profile.bin_centers, profile.n_macroparticles)
    plt.figure('beamInSlice')
    plt.plot(profile.Beam.dt[::100], profile.Beam.dE[::100], 'b.')
# ------------------------------------------------------------------------
# REMATCH THE BUNCHES WITH INTENSITY EFFECTS
# ------------------------------------------------------------------------
    if TotalInducedVoltage is not None:
        #        TotalInducedVoltage.profile.Beam.dt[:len(beam.dt)] = beam.dt
        #        TotalInducedVoltage.profile.Beam.dE[:len(beam.dE)] = beam.dE
        print('Applying intensity effects ...')
        for it in range(n_iterations):
            conv = 0.
            # Compute the induced voltage/potential for all the beam
            profile.track()
            TotalInducedVoltage.induced_voltage_sum()

            plt.figure('profile before induced voltage')
            profile.track()
            plt.plot(profile.bin_centers, profile.n_macroparticles)
#
#            plt.figure('inducedvoltage before induced voltage')
#            profile.track()
#            plt.plot(TotalInducedVoltage.time_array,TotalInducedVoltage.induced_voltage)
#
            induced_voltage_coordinates = TotalInducedVoltage.time_array
            induced_voltage = TotalInducedVoltage.induced_voltage
            induced_potential = - normalization_potential * cumtrapz(induced_voltage, dx=induced_voltage_coordinates[1] - induced_voltage_coordinates[0], initial=0)

            plt.figure('testInducedVolt')
            plt.plot(induced_voltage_coordinates, induced_voltage)
            plt.figure('testInducedPot')
            plt.plot(induced_voltage_coordinates, induced_potential)

            FullRingAndRF.potential_well_generation(n_points=n_points_potential,
                                                    dt_margin_percent=dt_margin_percent,
                                                    main_harmonic_option=main_harmonic_option)

            # Restrict the potential well inside the separatrix and put min on 0
            potential_well_coordinates, potential_well = potential_well_cut(
                FullRingAndRF.potential_well_coordinates,
                FullRingAndRF.potential_well)
            potential_well = potential_well - np.min(potential_well)

            temporary_beam = Beam(GeneralParameters, n_macro_per_bunch, intensity_per_bunch)
            for indexBatch in range(n_batch):
                for indexBunch in range(n_bunches):
                    # Extract the induced potential for the specific bucket
                    induced_potential_bunch = np.interp(potential_well_coordinates
                                                        + indexBunch * bunch_spacing_buckets * bucket_size_tau
                                                        + indexBatch * (batch_spacing_buckets + (n_bunches - 1) * bunch_spacing_buckets) * bucket_size_tau,
                                                        induced_voltage_coordinates, induced_potential)

                    # Recompute the phase space distribution for the new
                    # perturbed potential (containing induced_potential_bunch)
                    match_a_bunch(normalization_DeltaE, temporary_beam,
                                  potential_well_coordinates,
                                  potential_well + induced_potential_bunch, seed,
                                  distribution_options,
                                  full_ring_and_RF=FullRingAndRF)

                    dt = temporary_beam.dt
                    dE = temporary_beam.dE

                    # Compute RMS emittance to observe convergence
                    conv += np.pi * np.std(dt) * np.std(dE)

                    length_dt = len(dt)
                    length_dE = len(dE)
                    beam.dt[(indexBunch + n_bunches * indexBatch) * length_dt:(indexBunch + n_bunches * indexBatch + 1) * length_dt] = dt + (indexBunch * bunch_spacing_buckets * bucket_size_tau) + indexBatch * (batch_spacing_buckets + (n_bunches - 1) * bunch_spacing_buckets) * bucket_size_tau
                    beam.dE[(indexBunch + n_bunches * indexBatch) * length_dE:(indexBunch + n_bunches * indexBatch + 1) * length_dE] = dE

            print('iteration ' + str(it) + ', average RMS emittance (4sigma) = ' + str(4 * conv / n_bunches))
            profile.track()
            TotalInducedVoltage.induced_voltage_sum()


def compute_X_grid(normalization_DeltaE, time_array, potential_well,
                   distribution_variable):

    # Delta Energy array
    max_DeltaE = np.sqrt(np.max(potential_well) / normalization_DeltaE)
    coord_array_DeltaE = np.linspace(-float(max_DeltaE), float(max_DeltaE), len(time_array))

    # Resolution in time and energy
    time_resolution = time_array[1] - time_array[0]
    energy_resolution = coord_array_DeltaE[1] - coord_array_DeltaE[0]

    # Grid
    time_grid, deltaE_grid = np.meshgrid(time_array, coord_array_DeltaE)
    potential_well_grid = np.meshgrid(potential_well, potential_well)[0]
    H_grid = normalization_DeltaE * deltaE_grid**2 + potential_well_grid

    # Compute the action J
    J_array = np.zeros(shape=potential_well.shape, dtype=float)
    for i in range(len(J_array)):
        DELTA = np.sqrt((potential_well[i] - potential_well)[potential_well <= potential_well[i]] / normalization_DeltaE)
        J_array[i] = 1. / np.pi * np.trapz(DELTA, dx=time_array[1] - time_array[0])

    # Compute J grid
    sorted_H = potential_well[potential_well.argsort()]
    sorted_J = J_array[potential_well.argsort()]

    if distribution_variable == 'Action':
        J_grid = np.interp(H_grid, sorted_H, sorted_J,
                           left=0, right=np.inf)
        return sorted_H, sorted_J, J_grid, time_grid, deltaE_grid,\
            time_resolution, energy_resolution
    else:
        return sorted_H, sorted_J, H_grid, time_grid, deltaE_grid,\
            time_resolution, energy_resolution


def compute_H0(emittance, H, J):
    #  Estimation of H corresponding to the emittance
    return np.interp(emittance / (2. * np.pi), J, H)


def match_a_bunch(normalization_DeltaE, beam, potential_well_coordinates,
                  potential_well, seed, distribution_options,
                  full_ring_and_RF=None):

    if 'type' in distribution_options:
        distribution_type = distribution_options['type']
    else:
        distribution_type = None

    if 'exponent' in distribution_options:
        distribution_exponent = distribution_options['exponent']
    else:
        distribution_exponent = None

    if 'emittance' in distribution_options:
        emittance = distribution_options['emittance']
    else:
        emittance = None

    if 'bunch_length' in distribution_options:
        bunch_length = distribution_options['bunch_length']
    else:
        bunch_length = None

    if 'bunch_length_fit' in distribution_options:
        bunch_length_fit = distribution_options['bunch_length_fit']
    else:
        bunch_length_fit = None

    if 'density_variable' in distribution_options:
        distribution_variable = distribution_options['density_variable']
    else:
        distribution_variable = 'Hamiltonian'

    H, J, X_grid, time_grid, deltaE_grid, time_resolution, energy_resolution =\
        compute_X_grid(normalization_DeltaE, potential_well_coordinates,
                       potential_well, distribution_variable)

    # Choice of either H or J as the variable used
    if distribution_variable == 'Action':
        sorted_X = J
    elif distribution_variable == 'Hamiltonian':
        sorted_X = H
    else:
        # DistributionError
        raise SystemError('distribution_variable should be Action or Hamiltonian')

    if bunch_length is not None:
        n_points_grid = X_grid.shape[0]
        X0 = X0_from_bunch_length(bunch_length, bunch_length_fit, X_grid, sorted_X,
                                  n_points_grid, potential_well_coordinates,
                                  distribution_function, distribution_type,
                                  distribution_exponent, beam, full_ring_and_RF)
    elif emittance is not None:
        X0 = compute_H0(emittance, H, J)
    else:
        # DistributionError
        raise SystemError('You should specify either bunch_length or emittance')

    distribution = distribution_function(X_grid, distribution_type, X0, exponent=distribution_exponent)
    distribution[X_grid > np.max(H)] = 0
    distribution = distribution / np.sum(distribution)

    profile = np.sum(distribution, axis=0)

    return (time_grid, deltaE_grid, distribution, time_resolution,
            energy_resolution, profile)



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public License version 3 (GPL Version 3),
# copied verbatim in the file LICENSE.md.
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Function(s) for pre-processing input data**

:Authors: **Helga Timko**, **Alexandre Lasheen**, **Danilo Quartullo**,
    **Simon Albright**
'''

from __future__ import division, annotations

from builtins import range, str

import matplotlib.pyplot as plt
import numpy as np
from scipy.constants import c
from scipy.interpolate import splev, splrep, Akima1DInterpolator
from typing import TYPE_CHECKING

from ..plots.plot import fig_folder

if TYPE_CHECKING:
    from typing import List, Iterable, Tuple


class RingOptions:
    r""" Class to preprocess the synchronous data for Ring, interpolating it to
    every turn.

    Parameters
    ----------
    interpolation : str
        Interpolation options for the data points. Available options are
        'linear' (default), 'cubic', 'derivative' and 'akima'
    smoothing : float
        Smoothing value for 'cubic' interpolation
    flat_bottom : int
        Number of turns to be added on flat bottom; default is 0. Constant
        extrapolation is used for the synchronous data
    flat_top : int
        Number of turns to be added on flat top; default is 0. Constant
        extrapolation is used for the synchronous data
    t_start : int
        Starting index from which the time array input should be taken into
        account; default is 0
    t_end : int
        Last index up to which the time array input should be taken into
        account; default is -1
    plot : bool
        Option to plot interpolated arrays; default is False
    figdir : str
        Directory to save optional plot; default is 'fig'
    figname : str
        Figure name to save optional plot; default is 'preprocess_ramp'
    sampling : int
        Decimation value for plotting; default is 1

    """

    def __init__(self, interpolation='linear', smoothing=0, flat_bottom=0,
                 flat_top=0, t_start=None, t_end=None, plot=False,
                 figdir='fig', figname='preprocess_ramp', sampling=1):

        if interpolation in ['linear', 'cubic', 'derivative', 'akima']:
            self.interpolation = str(interpolation)
        else:
            # InputDataError
            raise RuntimeError("ERROR: Interpolation scheme in " +
                               "PreprocessRamp not recognised. Aborting...")

        self.smoothing = float(smoothing)

        if flat_bottom < 0:
            # MomentumError
            raise RuntimeError("ERROR: flat_bottom value in PreprocessRamp" +
                               " not recognised. Aborting...")
        else:
            self.flat_bottom = int(flat_bottom)

        if flat_top < 0:
            # MomentumError
            raise RuntimeError("ERROR: flat_top value in PreprocessRamp" +
                               " not recognised. Aborting...")
        else:
            self.flat_top = int(flat_top)

        self.t_start = t_start
        self.t_end = t_end

        if (plot is True) or (plot is False):
            self.plot = bool(plot)
        else:
            # TypeError
            raise RuntimeError("ERROR: plot value in PreprocessRamp" +
                               " not recognised. Aborting...")

        self.figdir = str(figdir)
        self.figname = str(figname)
        if sampling > 0:
            self.sampling = int(sampling)
        else:
            # TypeError
            raise RuntimeError("ERROR: sampling value in PreprocessRamp" +
                               " not recognised. Aborting...")

    def reshape_data(self, input_data, n_turns, n_sections,
                     interp_time='t_rev', input_to_momentum=False,
                     synchronous_data_type='momentum', mass=None, charge=None,
                     circumference=None, bending_radius=None):
        r"""Checks whether the user input is consistent with the expectation
        for the Ring object. The possibilites are detailed in the documentation
        of the Ring object.


        Parameters
        ----------
        input_data : Ring.synchronous_data, Ring.alpha_0,1,2
            Main input data to reshape
        n_turns : Ring.n_turns
            Number of turns the simulation should be. Note that if
            the input_data is passed as a tuple it is expected that the
            input_data is a program. Hence, the number of turns may not
            correspond to the input one and will be overwritten
        n_sections : Ring.n_sections
            The number of sections of the ring. The simulation is stopped
            if the input_data shape does not correspond to the expected number
            of sections.
        interp_time : str or float or float array [n_turns+1]
            Optional : defines the time on which the program will be
            interpolated. If 't_rev' is passed and if the input_data is
            momentum (see input_to_momentum option) the momentum program
            is interpolated on the revolution period (see preprocess()
            function). If a float or a float array is passed, the program
            is interpolated on that input ; default is 't_rev'
        input_to_momentum : bool
            Optional : flags if the input_data is the momentum program, the
            options defined below become necessary for conversion
        synchronous_data_type : str
            Optional : to be passed to the convert_data function if
            input_to_momentum ; default is 'momentum'
        mass : Ring.Particle.mass
            Optional : the mass of the particles in [eV/c**2] ; default is None
        charge : Ring.Particle.charge
            Optional : the charge of the particles in units of [e] ;
            default is None
        circumference : Ring.circumference
            Optional : the circumference of the ring ; default is None
        bending_radius : Ring.bending_radis
            Optional : the bending radius of magnets ; default is None

        Returns
        -------
        output_data
            Returns the data with the adequate shape for the Ring object

        """

        # TO BE IMPLEMENTED: if you pass a filename the function reads the file
        # and reshape the data
        if isinstance(input_data, str):
            pass

        # If single float, expands the value to match the input number of turns
        # and sections
        if isinstance(input_data, float) or isinstance(input_data, int):
            input_data = float(input_data)
            if input_to_momentum:
                input_data = convert_data(input_data, mass, charge,
                                          synchronous_data_type,
                                          bending_radius)
            output_data = input_data * np.ones((n_sections, n_turns + 1))

        # If tuple, separate time and synchronous data and check data
        elif isinstance(input_data, tuple):

            output_data = []

            # If there is only one section, it is expected that the user passes
            # a tuple with (time, data). However, the user can also pass a
            # tuple which size is the number of section as ((time, data), ).
            # and this if condition takes this into account
            if (n_sections == 1) and (len(input_data) > 1):
                input_data = (input_data, )

            if len(input_data) != n_sections:
                # InputDataError
                raise RuntimeError("ERROR in Ring: the input data " +
                                   "does not match the number of sections")

            # Loops over all the sections to interpolate the programs, appends
            # the results on the output_data list which is afterwards
            # converted to a numpy.array
            for index_section in range(n_sections):
                input_data_time = input_data[index_section][0]
                input_data_values = input_data[index_section][1]

                if input_to_momentum:
                    input_data_values = convert_data(input_data_values, mass,
                                                     charge,
                                                     synchronous_data_type,
                                                     bending_radius)

                if len(input_data_time) \
                        != len(input_data_values):
                    # InputDataError
                    raise RuntimeError("ERROR in Ring: synchronous data " +
                                       "does not match the time data")

                if input_to_momentum and (interp_time == 't_rev'):
                    output_data.append(self.preprocess(
                        mass,
                        circumference,
                        input_data_time,
                        input_data_values)[1])

                elif isinstance(interp_time, float) or \
                        isinstance(interp_time, int):
                    interp_time = float(interp_time)
                    interp_time = np.arange(
                        input_data_time[0],
                        input_data_time[-1],
                        interp_time)

                    output_data.append(np.interp(
                        interp_time,
                        input_data_time,
                        input_data_values))

                elif isinstance(interp_time, np.ndarray):
                    output_data.append(np.interp(
                        interp_time,
                        input_data_time,
                        input_data_values))

            output_data = np.array(output_data, ndmin=2, dtype=float)

        # If array/list, compares with the input number of turns and
        # if synchronous_data is a single value converts it into a (n_turns+1)
        # array
        elif isinstance(input_data, np.ndarray) or \
                isinstance(input_data, list):

            input_data = np.array(input_data, ndmin=2, dtype=float)

            if input_to_momentum:
                input_data = convert_data(input_data, mass, charge,
                                          synchronous_data_type,
                                          bending_radius)

            output_data = np.zeros((n_sections, n_turns + 1), dtype=float)

            # If the number of points is exactly the same as n_rf, this means
            # that the rf program for each harmonic is constant, reshaping
            # the array so that the size is [n_sections,1] for successful
            # reshaping
            if input_data.size == n_sections:
                input_data = input_data.reshape((n_sections, 1))

            if len(input_data) != n_sections:
                # InputDataError
                raise RuntimeError("ERROR in Ring: the input data " +
                                   "does not match the number of sections")

            for index_section in range(len(input_data)):
                if len(input_data[index_section]) == 1:
                    output_data[index_section] = input_data[index_section] * \
                        np.ones(n_turns + 1)

                elif len(input_data[index_section]) == (n_turns + 1):
                    output_data[index_section] = np.array(
                        input_data[index_section])

                else:
                    # InputDataError
                    raise RuntimeError("ERROR in Ring: The input data " +
                                       "does not match the proper length " +
                                       "(n_turns+1)")

        return output_data

    def preprocess(self, mass, circumference, time, momentum):
        r"""Function to pre-process acceleration ramp data, interpolating it to
        every turn. Currently it works only if the number of RF sections is
        equal to one, to be extended for multiple RF sections.

        Parameters
        ----------
        mass : float
            Particle mass [eV]
        circumference : float
            Ring circumference [m]
        time : float array
            Time points [s] corresponding to momentum data
        momentum : float array
            Particle momentum [eV/c]

        Returns
        -------
        float array
            Cumulative time [s]
        float array
            Interpolated momentum [eV/c]

        """

        # Some checks on the options
        if ((self.t_start is not None) and (self.t_start < time[0])) or \
                ((self.t_end is not None) and (self.t_end > time[-1])):
            # InputDataError
            raise RuntimeError("ERROR: [t_start, t_end] should be " +
                               "included in the passed time array.")

        # Obtain flat bottom data, extrapolate to constant
        beta_0 = np.sqrt(1 / (1 + (mass / momentum[0])**2))
        T0 = circumference / (beta_0 * c)  # Initial revolution period [s]
        shift = time[0] - self.flat_bottom * T0
        time_interp = shift + T0 * np.arange(0, self.flat_bottom + 1)
        beta_interp = beta_0 * np.ones(self.flat_bottom + 1)
        momentum_interp = momentum[0] * np.ones(self.flat_bottom + 1)

        time_interp = time_interp.tolist()
        beta_interp = beta_interp.tolist()
        momentum_interp = momentum_interp.tolist()

        time_start_ramp = np.max(time[momentum == momentum[0]])
        time_end_ramp = np.min(time[momentum == momentum[-1]])

        # Interpolate data recursively
        if self.interpolation == 'linear':

            interps = self._linear_interpolation(time_interp, momentum_interp,
                                                 beta_interp, circumference,
                                                 time, momentum, mass)
            time_interp, momentum_interp, beta_interp = interps

        elif self.interpolation == 'cubic':

            interps = self._cubic_interpolation(time_interp, momentum_interp,
                                                beta_interp, circumference,
                                                time, momentum, mass,
                                                time_start_ramp, time_end_ramp)
            time_interp, momentum_interp, beta_interp = interps

        # Interpolate momentum in 1st derivative to maintain smooth B-dot
        elif self.interpolation == 'derivative':

            interps = self._derivative_interpolation(time_interp,
                                                     momentum_interp,
                                                     beta_interp,
                                                     circumference, time,
                                                     momentum, mass)
            time_interp, momentum_interp, beta_interp = interps

        elif self.interpolation == 'akima':

            interps = self._akima_interpolation(time_interp, momentum_interp,
                                                beta_interp, circumference,
                                                time, momentum, mass)
            time_interp, momentum_interp, beta_interp = interps

        time_interp.pop()
        time_interp = np.asarray(time_interp)
        beta_interp = np.asarray(beta_interp)
        momentum_interp = np.asarray(momentum_interp)

        # Obtain flat top data, extrapolate to constant
        if self.flat_top > 0:
            time_interp = np.append(
                time_interp,
                time_interp[-1] + circumference * np.arange(1, self.flat_top + 1)
                / (beta_interp[-1] * c))

            beta_interp = np.append(
                beta_interp, beta_interp[-1] * np.ones(self.flat_top))

            momentum_interp = np.append(
                momentum_interp,
                momentum_interp[-1] * np.ones(self.flat_top))

        # Cutting the input momentum on the desired cycle time
        if self.t_start is not None:
            initial_index = np.min(np.where(time_interp >= self.t_start)[0])
        else:
            initial_index = 0
        if self.t_end is not None:
            final_index = np.max(np.where(time_interp <= self.t_end)[0]) + 1
        else:
            final_index = len(time_interp)
        time_interp = time_interp[initial_index:final_index]
        momentum_interp = momentum_interp[initial_index:final_index]

        if self.plot:
            # Directory where longitudinal_plots will be stored
            fig_folder(self.figdir)

            # Plot
            plt.figure(1, figsize=(8, 6))
            ax = plt.axes([0.15, 0.1, 0.8, 0.8])
            ax.plot(time_interp[::self.sampling],
                    momentum_interp[::self.sampling],
                    label='Interpolated momentum')
            ax.plot(time, momentum, '.', label='input momentum', color='r',
                    markersize=0.5)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel("p [eV]")
            ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            ax.legend = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                                   ncol=2, mode="expand", borderaxespad=0.)

            # Save figure
            fign = self.figdir + '/preprocess_' + self.figname
            plt.savefig(fign)
            plt.clf()

        return time_interp, momentum_interp

    def _linear_interpolation(self, time_interp: List[float],
                              momentum_interp: List[float],
                              beta_interp: List[float], circumference: float,
                              time: Iterable[float], momentum: Iterable[float],
                              mass: float) -> Tuple[Iterable[float], ...]:

        time_interp.append(time_interp[-1]
                            + circumference / (beta_interp[0] * c))

        i = self.flat_bottom
        for k in range(1, len(time)):

            while time_interp[i + 1] <= time[k]:

                momentum_interp.append(
                    momentum[k - 1] + (momentum[k] - momentum[k - 1]) *
                    (time_interp[i + 1] - time[k - 1]) /
                    (time[k] - time[k - 1]))

                next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                            mass, circumference)
                beta_interp.append(next_beta)
                time_interp.append(time_interp[i + 1] + next_time)

                i += 1

        return time_interp, momentum_interp, beta_interp

    def _cubic_interpolation(self, time_interp: List[float],
                             momentum_interp: List[float],
                             beta_interp: List[float], circumference: float,
                             time: Iterable[float], momentum: Iterable[float],
                             mass: float, time_start_ramp: float,
                             time_end_ramp: float) -> Tuple[Iterable[float],
                                                            ...]:

        interp_funtion_momentum = splrep(
            time[(time >= time_start_ramp) * (time <= time_end_ramp)],
            momentum[(time >= time_start_ramp) * (time <= time_end_ramp)],
            s=self.smoothing)

        i = self.flat_bottom

        time_interp.append(
            time_interp[-1] + circumference / (beta_interp[0] * c))

        while time_interp[i] <= time[-1]:

            if (time_interp[i + 1] < time_start_ramp):

                momentum_interp.append(momentum[0])
                next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                            mass, circumference)

                beta_interp.append(next_beta)
                time_interp.append(time_interp[i + 1] + next_time)

            elif (time_interp[i + 1] > time_end_ramp):

                momentum_interp.append(momentum[-1])

                next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                            mass, circumference)
                beta_interp.append(next_beta)
                time_interp.append(time_interp[i + 1] + next_time)

            else:

                momentum_interp.append(
                    splev(time_interp[i + 1], interp_funtion_momentum))
                next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                            mass, circumference)
                beta_interp.append(next_beta)
                time_interp.append(time_interp[i + 1] + next_time)

            i += 1

        return time_interp, momentum_interp, beta_interp

    def _derivative_interpolation(self, time_interp: List[float],
                                  momentum_interp: List[float],
                                  beta_interp: List[float],
                                  circumference: float,
                                  time: Iterable[float],
                                  momentum: Iterable[float],
                                  mass: float) -> Tuple[Iterable[float], ...]:

        momentum_initial = momentum_interp[0]
        momentum_derivative = np.gradient(momentum) / np.gradient(time)

        momentum_derivative_interp = [0] * self.flat_bottom + \
            [momentum_derivative[0]]
        integral_point = momentum_initial

        i = self.flat_bottom

        time_interp.append(
            time_interp[-1] + circumference / (beta_interp[0] * c))

        while time_interp[i] <= time[-1]:

            derivative_point = np.interp(time_interp[i + 1], time,
                                            momentum_derivative)
            momentum_derivative_interp.append(derivative_point)
            integral_point += (time_interp[i + 1] - time_interp[i]) \
                * derivative_point

            momentum_interp.append(integral_point)
            next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                        mass, circumference)
            beta_interp.append(next_beta)
            time_interp.append(time_interp[i + 1] + next_time)

            i += 1

        # Adjust result to get flat top energy correct as derivation and
        # integration leads to ~10^-8 error in flat top momentum
        momentum_interp = np.asarray(momentum_interp)
        momentum_interp -= momentum_interp[0]
        momentum_interp /= momentum_interp[-1]
        momentum_interp *= momentum[-1] - momentum[0]

        momentum_interp += momentum[0]

        return time_interp, momentum_interp, beta_interp

    def _akima_interpolation(self, time_interp: List[float],
                             momentum_interp: List[float],
                             beta_interp: List[float],
                             circumference: float,
                             time: Iterable[float],
                             momentum: Iterable[float],
                             mass: float) -> Tuple[Iterable[float], ...]:


        interp_func = Akima1DInterpolator(time, momentum)

        i = self.flat_bottom

        time_interp.append(
            time_interp[-1] + circumference / (beta_interp[0] * c))

        while time_interp[i] <= time[-1]:

            momentum_interp.append(interp_func(time_interp[i + 1]))
            next_time, next_beta = self._next_time_beta(momentum_interp[i+1],
                                                        mass, circumference)
            beta_interp.append(next_beta)
            time_interp.append(time_interp[i + 1] + next_time)

            i += 1

        return time_interp, momentum_interp, beta_interp


    def _next_time_beta(self, next_momentum: float, mass: float,
                        circumference: float):

        beta = np.sqrt(1 / (1 + (mass / next_momentum)**2))

        time = circumference / (beta*c)

        return time, beta


def convert_data(synchronous_data, mass, charge,
                 synchronous_data_type='momentum', bending_radius=None):
    """ Function to convert synchronous data (i.e. energy program of the
    synchrotron) into momentum.

    Parameters
    ----------
    synchronous_data : float array
        The synchronous data to be converted to momentum
    mass : float or Particle.mass
        The mass of the particles in [eV/c**2]
    charge : int or Particle.charge
        The charge of the particles in units of [e]
    synchronous_data_type : str
        Type of input for the synchronous data ; can be 'momentum',
        'total energy', 'kinetic energy' or 'bending field' (last case
        requires bending_radius to be defined)
    bending_radius : float
        Bending radius in [m] in case synchronous_data_type is
        'bending field'

    Returns
    -------
    momentum : float array
        The input synchronous_data converted into momentum [eV/c]

    """

    if synchronous_data_type == 'momentum':
        momentum = synchronous_data
    elif synchronous_data_type == 'total energy':
        momentum = np.sqrt(synchronous_data**2 - mass**2)
    elif synchronous_data_type == 'kinetic energy':
        momentum = np.sqrt((synchronous_data + mass)**2 - mass**2)
    elif synchronous_data_type == 'bending field':
        if bending_radius is None:
            # InputDataError
            raise RuntimeError("ERROR in Ring: bending_radius is not " +
                               "defined and is required to compute " +
                               "momentum")
        momentum = synchronous_data * bending_radius * charge * c
    else:
        # InputDataError
        raise RuntimeError("ERROR in Ring: Synchronous data" +
                           " type not recognized!")

    return momentum


def load_data(filename, ignore=0, delimiter=None):
    r"""Helper function to load column-by-column data from a txt file to numpy
    arrays.

    Parameters
    ----------
    filename : str
        Name of the file containing the data.
    ignore : int
        Number of lines to ignore from the head of the file.
    delimiter : str
        Delimiting character between columns.

    Returns
    -------
    list of arrays
        Input data, column by column.

    """

    data = np.loadtxt(str(filename), skiprows=int(ignore),
                      delimiter=str(delimiter))

    return [np.ascontiguousarray(data[:, i]) for i in range(len(data[0]))]



# coding: utf8
# Copyright 2014-2020 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Scaling of longitudinal beam and machine parameters, with user interface.**

:Authors: **Konstantinos Iliakis**, **Helga Timko**
'''

import numpy as np
from PyQt5 import QtCore, QtWidgets
from scipy import integrate
from scipy.constants import c, e, m_p

# Machine-dependent parameters [SI-units] -------------------------------------
set_ups = {'PSB': '0',
           'CPS': '1',
           'SPS, Q20': '2', 'SPS, Q22': '3', 'SPS, Q26': '4',
           'LHC, -2016': '5', 'LHC, 2017-': '6'}
gamma_ts = {'0': 4.0767,
            '1': np.sqrt(37.2),
            '2': 18., '3': 20., '4': 22.83,
            '5': 55.759505, '6': 53.8}
harmonics = {'0': 1,
             '1': 21,
             '2': 4620, '3': 4620, '4': 4620,
             '5': 35640, '6': 35640}
circumferences = {'0': 2 * np.pi * 25,
                  '1': 2 * np.pi * 100.,
                  '2': 2 * np.pi * 1100.009, '3': 2 * np.pi * 1100.009, '4': 2 * np.pi * 1100.009,
                  '5': 26658.883, '6': 26658.883}
energies_fb = {'0': (160.e6 + m_p * c**2 / e),
               '1': (2.0e9 + m_p * c**2 / e),
               '2': 25.92e9, '3': 25.92e9, '4': 25.92e9,
               '5': 450.e9, '6': 450.e9}
energies_ft = {'0': (2.0e9 + m_p * c**2 / e),
               '1': 25.92e9,
               '2': 450.e9, '3': 450.e9, '4': 450.e9,
               '5': 6.5e12, '6': 6.5e12}
# Machine-dependent parameters [SI-units] -------------------------------------


class ParameterScaling:

    @property
    def phi_b(self):
        return self.omega_rf * self.tau / 2.

    @property
    def delta_b(self):
        return self.dE_b / (self.beta_sq * self.energy)

    @property
    def dE_b(self):
        return np.sqrt(self.beta_sq * self.energy * self.voltage * (1 -
                       np.cos(self.phi_b)) / (np.pi * self.harmonic * self.eta_0))

    @property
    def integral(self):
        return integrate.quad(lambda x: np.sqrt(2. * (np.cos(x) -
                              np.cos(self.phi_b))), 0, self.phi_b)[0]

    @property
    def emittance(self):
        return 4. * self.energy * self.omega_s0 * self.beta_sq * self.integral / \
            (self.omega_rf**2 * self.eta_0)

    def relativistic_quantities(self):

        self.momentum = np.sqrt(self.energy**2 - self.mass**2)
        self.tb1.append(f"    Synchronous momentum: {self.momentum} eV")

        self.kinetic_energy = self.energy - self.mass
        self.tb1.append(f"    Synchronous kinetic energy: {self.kinetic_energy} eV")

        self.gamma = self.energy / self.mass
        self.tb1.append(f"    Synchronous relativistic gamma: {self.gamma}")

        self.beta = np.sqrt(1. - 1. / self.gamma**2)
        self.tb1.append(f"    Synchronous relativistic beta: {self.beta}")

        self.beta_sq = self.beta ** 2
        self.tb1.append(f"    Synchronous relativistic beta squared: {self.beta_sq}\n")

    def frequencies(self):

        self.t_rev = self.circumference / (self.beta * c)
        self.tb1.append(f"    Revolution period: {self.t_rev * 1.e6} us")

        self.f_rev = 1. / self.t_rev
        self.tb1.append(f"    Revolution frequency: {self.f_rev} Hz")

        self.omega_rev = 2. * np.pi * self.f_rev
        self.tb1.append(f"        Angular revolution frequency: {self.omega_rev} 1/s")

        self.f_RF = self.harmonic * self.f_rev
        self.tb1.append(f"    RF frequency: {self.f_RF * 1.e-6} MHz")

        self.omega_rf = 2. * np.pi * self.f_RF
        self.tb1.append(f"        Angular RF frequency: {self.omega_rf} 1/s\n")

    def tune(self):

        self.eta_0 = np.fabs(1. / self.gamma_t**2 - 1. / self.gamma**2)
        self.tb1.append(f"    Slippage factor (zeroth order): {self.eta_0}")

        self.Q_s0 = np.sqrt(self.harmonic * self.voltage * self.eta_0 /
                            (2. * np.pi * self.beta_sq * self.energy))
        self.tb1.append(f"    Central synchrotron tune: {self.Q_s0}")

        self.f_s0 = self.Q_s0 * self.f_rev
        self.tb1.append(f"    Central synchrotron frequency: {self.f_s0}")

        self.omega_s0 = 2. * np.pi * self.f_s0
        self.tb1.append(f"        Angular synchrotron frequency: {self.omega_s0} 1/s\n")

    def bucket_parameters(self):

        self.tb1.append("Bucket parameters assume: single RF, stationary case, and no intensity effects.\n")

        self.bucket_area = 8. * np.sqrt(2. * self.beta_sq * self.energy * self.voltage /
                                        (np.pi * self.harmonic * self.eta_0)) / self.omega_rf
        self.tb1.append(f"    Bucket area: {self.bucket_area} eVs")

        self.dt_max = 0.5 * self.t_rev / self.harmonic
        self.tb1.append(f"    Half of bucket length: {self.dt_max * 1.e9} ns")

        self.dE_max = np.sqrt(2. * self.beta**2 * self.energy * self.voltage /
                              (np.pi * self.eta_0 * self.harmonic))
        self.tb1.append(f"    Half of bucket height: {self.dE_max * 1.e-6} MeV")

        self.delta_max = self.dE_max / (self.beta_sq * self.energy)
        self.tb1.append(f"        In relative momentum offset: {self.delta_max}\n")

    def emittance_from_bunch_length(self, four_sigma_bunch_length):

        self.tau = four_sigma_bunch_length
        if self.tau >= 2. * self.dt_max:
            self.tb1.append("Chosen bunch length too large for this bucket. Aborting!")
            raise RuntimeError("Chosen bunch length too large for this bucket. Aborting!")
        self.tb1.append(f"Calculating emittance of 4-sigma bunch length: {self.tau * 1.e9} ns")
        self.tb1.append(f"    Emittance contour in phase: {self.phi_b} rad")
        self.tb1.append(f"    Emittance contour in relative momentum: {self.delta_b}")
        self.tb1.append(f"    Emittance contour in energy offset: {self.dE_b * 1.e-6} MeV")
        self.tb1.append(f"    R.m.s. bunch length is: {self.tau * c / 4 * 100} cm")
        self.tb1.append(f"    R.m.s. energy spread is: {0.5 * self.dE_b / self.kinetic_energy}")
        self.tb1.append(f"    Longitudinal emittance is: {self.emittance} eVs\n")

    def bunch_length_from_emittance(self, emittance):

        self.emittance_aim = emittance

        if self.emittance_aim >= self.bucket_area:
            self.tb1.append("Chosen emittance too large for this bucket. Aborting!")
            raise RuntimeError("Chosen emittance too large for this bucket. Aborting!")
        self.tb1.append(f"Calculating 4-sigma bunch length for an emittance of {self.emittance_aim} eVs")

        # Make a guess, iterate to get closer
        self.tau = self.dt_max / 2.
        while (np.fabs((self.emittance - self.emittance_aim)
                       / self.emittance_aim) > 0.001):
            self.tau *= np.sqrt(self.emittance_aim / self.emittance)

        self.tb1.append(f"    Bunch length is: {self.tau * 1.e9} ns")
        self.tb1.append(f"    Corresponding matched rms relative momentum offset: {self.delta_b}")
        self.tb1.append(f"    Emittance contour in phase: {self.phi_b} rad")

    def setupUi(self, mainWindow):

        mainWindow.setObjectName("mainWindow")
        mainWindow.resize(586, 611)
        mainWindow.setWindowOpacity(1.0)
        mainWindow.setFixedSize(mainWindow.size())

        # Label "Machine/Optics"
        self.lbMachine = QtWidgets.QLabel(mainWindow)
        self.lbMachine.setGeometry(QtCore.QRect(20, 20, 120, 17))
        self.lbMachine.setMinimumSize(QtCore.QSize(70, 0))
        self.lbMachine.setMaximumSize(QtCore.QSize(16777215, 17))
        self.lbMachine.setObjectName("lbMachine")

        # Label "Energy"
        self.lbEnergy = QtWidgets.QLabel(mainWindow)
        self.lbEnergy.setGeometry(QtCore.QRect(20, 80, 70, 17))
        self.lbEnergy.setObjectName("lbEnergy")
        # Custom energy box
        self.leCustom = QtWidgets.QLineEdit(mainWindow)
        self.leCustom.setEnabled(True)
        self.leCustom.setGeometry(QtCore.QRect(145, 100, 70, 25))
        self.leCustom.setStyleSheet("background-color: rgb(255, 255, 255);\n"
                                    "border-color: rgb(0, 0, 0);")
        self.leCustom.hide()
        self.leCustom.setText("")
        self.leCustom.setObjectName("leCustom")
        # Custom energy label (unit)
        self.lbEV1 = QtWidgets.QLabel(mainWindow)
        self.lbEV1.setEnabled(True)
        self.lbEV1.setGeometry(QtCore.QRect(220, 100, 30, 25))
        self.lbEV1.setObjectName("lbEV1")
        self.lbEV1.hide()

        # Label "Gamma Transition"
        self.rbGammaT = QtWidgets.QLabel(mainWindow)
        self.rbGammaT.setGeometry(QtCore.QRect(260, 80, 120, 17))
        self.rbGammaT.setObjectName("rbGammaT")
        # Custom gamma_t box
        self.reCustom = QtWidgets.QLineEdit(mainWindow)
        self.reCustom.setEnabled(True)
        self.reCustom.setGeometry(QtCore.QRect(385, 100, 70, 25))
        self.reCustom.setStyleSheet("background-color: rgb(255, 255, 255);\n"
                                    "border-color: rgb(0, 0, 0);")
        self.reCustom.hide()
        self.reCustom.setText("")
        self.reCustom.setObjectName("reCustom")

        # Label "Voltage" with units
        self.lbVoltage = QtWidgets.QLabel(mainWindow)
        self.lbVoltage.setGeometry(QtCore.QRect(20, 160, 70, 25))
        self.lbVoltage.setObjectName("lbVoltage")
        self.lbEV2 = QtWidgets.QLabel(mainWindow)
        self.lbEV2.setGeometry(QtCore.QRect(150, 160, 31, 25))
        self.lbEV2.setObjectName("lbEV2")
        self.leVoltage = QtWidgets.QLineEdit(mainWindow)
        self.leVoltage.setGeometry(QtCore.QRect(80, 155, 70, 25))
        self.leVoltage.setStyleSheet("background-color: rgb(255, 255, 255);\n"
                                     "border-color: rgb(0, 0, 0);")
        self.leVoltage.setText("")
        self.leVoltage.setObjectName("leVoltage")

        # Label "Optional"
        self.lbOptional = QtWidgets.QLabel(mainWindow)
        self.lbOptional.setGeometry(QtCore.QRect(20, 230, 70, 17))
        self.lbOptional.setObjectName("lbOptional")

        # Label "Emittance" with units
        self.leEmittance = QtWidgets.QLineEdit(mainWindow)
        self.leEmittance.setGeometry(QtCore.QRect(130, 270, 70, 25))
        self.leEmittance.setStyleSheet("background-color: rgb(255, 255, 255);\n"
                                       "border-color: rgb(0, 0, 0);")
        self.leEmittance.setText("")
        self.leEmittance.setObjectName("leEmittance")
        self.lbEVS1 = QtWidgets.QLabel(mainWindow)
        self.lbEVS1.setGeometry(QtCore.QRect(200, 275, 41, 25))
        self.lbEVS1.setObjectName("lbEVS1")
        self.lbEVS2 = QtWidgets.QLabel(mainWindow)
        self.lbEVS2.setGeometry(QtCore.QRect(330, 275, 41, 25))
        self.lbEVS2.setObjectName("lbEVS2")

        # Label "Bunch Length" with units
        self.leBunchLength = QtWidgets.QLineEdit(mainWindow)
        self.leBunchLength.setGeometry(QtCore.QRect(260, 270, 70, 25))
        self.leBunchLength.setStyleSheet("background-color: rgb(255, 255, 255);\n"
                                         "border-color: rgb(0, 0, 0);")
        self.leBunchLength.setText("")
        self.leBunchLength.setObjectName("leBunchLength")

        # "Submit" button
        self.pbSubmit = QtWidgets.QPushButton(mainWindow)
        self.pbSubmit.setGeometry(QtCore.QRect(230, 320, 101, 27))
        self.pbSubmit.setObjectName("pbSumbit")
        self.tb1 = QtWidgets.QTextBrowser(mainWindow)
        self.tb1.setGeometry(QtCore.QRect(10, 350, 561, 241))
        self.tb1.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.tb1.setObjectName("tb1")

        # Drop-down menus Machine/Optics, Energy, Gamma Transition
        self.cbMachine = QtWidgets.QComboBox(mainWindow)
        self.cbMachine.setGeometry(QtCore.QRect(20, 40, 115, 25))
        self.cbMachine.setEditable(False)
        self.cbMachine.setObjectName("cbMachine")
        for i in range(len(gamma_ts)):
            self.cbMachine.addItem("")
        self.cbEnergy = QtWidgets.QComboBox(mainWindow)
        self.cbEnergy.setGeometry(QtCore.QRect(20, 100, 115, 25))
        self.cbEnergy.setObjectName("cbEnergy")
        self.cbEnergy.addItem("")
        self.cbEnergy.addItem("")
        self.cbEnergy.addItem("")
        self.cbGammaT = QtWidgets.QComboBox(mainWindow)
        self.cbGammaT.setGeometry(QtCore.QRect(260, 100, 115, 25))
        self.cbGammaT.setObjectName("cbGammaT")
        self.cbGammaT.addItem("")
        self.cbGammaT.addItem("")

        # Radio button Bunch Length
        self.rbBunchLength = QtWidgets.QRadioButton(mainWindow)
        self.rbBunchLength.setGeometry(QtCore.QRect(260, 250, 140, 22))
        self.rbBunchLength.setObjectName("rbBunchLength")

        # Radio button Emittance
        self.rbEmittance = QtWidgets.QRadioButton(mainWindow)
        self.rbEmittance.setGeometry(QtCore.QRect(130, 250, 100, 22))
        self.rbEmittance.setObjectName("rbEmittance")

        # Radio button No option
        self.rbNoOption = QtWidgets.QRadioButton(mainWindow)
        self.rbNoOption.setGeometry(QtCore.QRect(20, 250, 100, 22))
        self.rbNoOption.setObjectName('rbNoOption')
        self.rbNoOption.setChecked(True)

        self.retranslateUi(mainWindow)
        QtCore.QMetaObject.connectSlotsByName(mainWindow)
        self.addactions(mainWindow)

    def retranslateUi(self, mainWindow):

        _translate = QtCore.QCoreApplication.translate

        # Label texts
        mainWindow.setWindowTitle(_translate("mainWindow", "Bunch Parameter Calculator"))
        self.lbMachine.setText(_translate("mainWindow", "Machine, Optics"))
        self.lbEnergy.setText(_translate("mainWindow", "Energy"))
        self.lbEV1.setText(_translate("mainWindow", "[eV]"))
        self.rbGammaT.setText(_translate("mainWindow", "Transition Gamma"))
        self.lbVoltage.setText(_translate("mainWindow", "Voltage"))
        self.lbEV2.setText(_translate("mainWindow", "[V]"))
        self.lbOptional.setText(_translate("mainWindow", "Optional"))
        self.rbEmittance.setText(_translate("mainWindow", "Emittance"))
        self.lbEVS1.setText(_translate("mainWindow", "[eVs]"))
        self.lbEVS2.setText(_translate("mainWindow", "[s]"))
        self.rbBunchLength.setText(_translate("mainWindow", "Bunch Length"))
        self.rbNoOption.setText(_translate("mainWindow", "No Options"))
        self.pbSubmit.setText(_translate("mainWindow", "Submit"))

        # Options in roll-down menu
        for i, key in enumerate(set_ups.keys()):
            self.cbMachine.setItemText(i, _translate("mainWindow", key))
        self.cbEnergy.setItemText(0, _translate("mainWindow", "Flat bottom"))
        self.cbEnergy.setItemText(1, _translate("mainWindow", "Flat top"))
        self.cbEnergy.setItemText(2, _translate("mainWindow", "Custom"))
        self.cbGammaT.setItemText(0, _translate("mainWindow", "Default"))
        self.cbGammaT.setItemText(1, _translate("mainWindow", "Custom"))

    def addactions(self, mainWindow):

        self.pbSubmit.clicked.connect(self.pbHandler)
        self.cbEnergy.activated[str].connect(self.cbEnergyHandler)
        self.cbGammaT.activated[str].connect(self.cbGammaTHandler)

    def pbHandler(self):

        self.machine = str(self.cbMachine.currentText())
        self.setup = set_ups[self.machine]

        self.energy_type = self.cbEnergy.currentText()
        if self.energy_type == 'Custom':
            self.custom_energy = self.leCustom.text()
            try:
                self.energy = np.double(self.custom_energy)
            except ValueError:
                self.tb1.append("Energy not recognized!")
                return

        self.gamma_t = gamma_ts[self.setup]
        self.gamma_t_type = self.cbGammaT.currentText()
        if self.gamma_t_type == 'Custom':
            self.custom_gamma_t = self.reCustom.text()
            try:
                self.gamma_t = np.double(self.custom_gamma_t)
            except ValueError:
                self.tb1.append("Gamma transition not recognized!")
                return

        self.voltage = self.leVoltage.text()
        self.emittance_target = self.leEmittance.text()
        self.bunch_length_target = self.leBunchLength.text()

        self.tb1.append("\n\n" + "**************************** BEAM PARAMETER CALCULATOR ****************************" + "\n")
        self.tb1.append(f"Input -- chosen machine/optics: {self.machine}\n")

        # Derived parameters --------------------------------------------------

        self.alpha = 1. / self.gamma_t**2
        self.tb1.append(f"    * with relativistic gamma at transition: {self.gamma_t}")
        self.tb1.append(f"    * with momentum compaction factor: {self.alpha}")

        self.harmonic = harmonics[self.setup]
        self.tb1.append(f"    * with main harmonic: {self.harmonic}")

        self.circumference = circumferences[self.setup]
        self.tb1.append(f"    * and machine circumference: {self.circumference} m\n")

        if self.energy_type == 'Flat bottom':
            self.energy = energies_fb[self.setup]
        elif self.energy_type == 'Flat top':
            self.energy = energies_ft[self.setup]
        self.tb1.append(f"Input -- synchronous total energy: {self.energy * 1.e-6} MeV")

        try:
            self.voltage = np.double(self.voltage)
        except ValueError:
            self.tb1.append("Voltage not recognised!")
            return
        self.tb1.append(f"Input -- RF voltage: {self.voltage * 1.e-6} MV")

        self.mass = m_p * c**2 / e
        self.tb1.append(f"Input -- particle mass: {self.mass * 1.e-6} MeV\n")

        # Derived quantities --------------------------------------------------
        self.relativistic_quantities()
        self.frequencies()
        self.tune()
        self.bucket_parameters()

        if self.rbEmittance.isChecked():
            try:
                self.emittance_target = np.double(self.emittance_target)
            except ValueError:
                self.tb1.append("Target emittance not recognised!")
                return
            self.bunch_length_from_emittance(self.emittance_target)
        elif self.rbBunchLength.isChecked():
            try:
                self.bunch_length_target = np.double(self.bunch_length_target)
            except ValueError:
                self.tb1.append("Target bunch length not recognised!")
                return
            self.emittance_from_bunch_length(self.bunch_length_target)

    def cbEnergyHandler(self, text):
        if text == 'Custom':
            self.leCustom.show()
            self.lbEV1.show()
        else:
            self.leCustom.hide()
            self.lbEV1.hide()

    def cbGammaTHandler(self, text):
        if text == 'Custom':
            self.reCustom.show()
        else:
            self.reCustom.hide()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    mainWindow = QtWidgets.QMainWindow()
    ui = ParameterScaling()
    ui.setupUi(mainWindow)
    mainWindow.show()
    sys.exit(app.exec_())



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Logging messages**

:Authors: **Helga Timko**
'''

import logging


class Logger:
    """Class to log messages coming from other classes. Messages contain 
    {Time stamp} {Class name} {Log level} {Message}. Errors, warnings and info
    are logged into the console. To disable logging, call Logger().disable()

    Parameters
    ----------
    debug : bool
        Log DEBUG messages in 'debug.log'; default is False

    """

    def __init__(self, debug=False):

        # Root logger on DEBUG level
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)

        # Console handler on INFO level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        log_format = logging.Formatter(
            "%(asctime)s %(name)-25s %(levelname)-9s %(message)s")
        console_handler.setFormatter(log_format)
        root_logger.addHandler(console_handler)

        # File logger on DEBUG level
        if debug:
            file_handler = logging.FileHandler('debug.log', mode='w')
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(log_format)
            root_logger.addHandler(file_handler)

        logging.info("Start logging")
        if debug:
            logging.debug("Logger in debug mode")

    def disable(self):
        """Disables all logging."""

        logging.info("Disable logging")
        logging.disable(level=logging.NOTSET)




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to save beam statistics in h5 files**

:Authors: **Danilo Quartullo**, **Helga Timko**
'''


import h5py as hp
import numpy as np
import os

class BunchMonitor:

    ''' Class able to save bunch data into h5 file. Use 'buffer_time' to select 
        the frequency of saving to file in number of turns.
        If in the constructor a Profile object is passed, that means that one
        wants to save the gaussian-fit bunch length as well (obviously the 
        Profile object has to have the fit_option set to 'gaussian').
    '''

    def __init__(self, Ring, RFParameters, Beam, filename,
                 buffer_time=None,
                 Profile=None, PhaseLoop=None, LHCNoiseFB=None):

        self.filename = filename
        self.n_turns = Ring.n_turns
        self.i_turn = 0
        self.buffer_time = buffer_time
        if buffer_time is None:
            self.buffer_time = self.n_turns
        self.rf_params = RFParameters
        self.beam = Beam
        self.profile = Profile
        if self.profile:
            if self.profile.fit_option is not None:
                self.fit_option = True
            else:
                self.fit_option = False
        else:
            self.fit_option = False
        self.PL = PhaseLoop
        self.LHCNoiseFB = LHCNoiseFB

        # Initialise data and save initial state
        self.init_data(self.filename, (self.n_turns + 1,))

        # Track at initialisation
        self.track()

    def track(self):

        self.beam.statistics()

        # Write buffer with i_turn = RFcounter - 1
        self.write_buffer()

        # Synchronise to i_turn = RFcounter
        self.i_turn += 1

        if self.i_turn > 0 and (self.i_turn % self.buffer_time) == 0:
            self.open()
            self.write_data(self.h5file['Beam'], (self.n_turns + 1,))
            self.close()
            self.init_buffer()

    def init_data(self, filename, dims):

        # Prepare data
        self.beam.statistics()

        # create directory to save h5 file if needed
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

        # Open file
        self.h5file = hp.File(filename + '.h5', 'w')
        self.h5file.require_group('Beam')

        # Create datasets and write first data points
        h5group = self.h5file['Beam']

        h5group.create_dataset("n_macroparticles_alive", shape=dims,
                               dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["n_macroparticles_alive"][0] = self.beam.n_macroparticles_alive

        h5group.create_dataset("mean_dt", shape=dims, dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["mean_dt"][0] = self.beam.mean_dt

        h5group.create_dataset("mean_dE", shape=dims, dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["mean_dE"][0] = self.beam.mean_dE

        h5group.create_dataset("sigma_dt", shape=dims, dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["sigma_dt"][0] = self.beam.sigma_dt

        h5group.create_dataset("sigma_dE", shape=dims, dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["sigma_dE"][0] = self.beam.sigma_dE

        h5group.create_dataset("epsn_rms_l", shape=dims, dtype='f',
                               compression="gzip", compression_opts=9)
        h5group["epsn_rms_l"][0] = self.beam.epsn_rms_l

        if self.fit_option:

            h5group.create_dataset("bunch_length", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["bunch_length"][0] = self.profile.bunchLength

        if self.PL:

            h5group.create_dataset("PL_omegaRF", shape=dims,
                                   dtype=np.float64,
                                   compression="gzip", compression_opts=9)
            h5group["PL_omegaRF"][0] = self.rf_params.omega_rf[0, 0]

            h5group.create_dataset("PL_phiRF", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["PL_phiRF"][0] = self.rf_params.phi_rf[0, 0]

            h5group.create_dataset("PL_bunch_phase", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["PL_bunch_phase"][0] = self.PL.phi_beam

            h5group.create_dataset("PL_phase_corr", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["PL_phase_corr"][0] = self.PL.dphi

            h5group.create_dataset("PL_omegaRF_corr", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["PL_omegaRF_corr"][0] = self.PL.domega_rf

            h5group.create_dataset("SL_dphiRF", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["SL_dphiRF"][0] = self.rf_params.dphi_rf[0]

            h5group.create_dataset("RL_drho", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["RL_drho"][0] = self.PL.drho

        if self.LHCNoiseFB:

            h5group.create_dataset("LHC_noise_FB_factor", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["LHC_noise_FB_factor"][0] = self.LHCNoiseFB.x

            h5group.create_dataset("LHC_noise_FB_bl", shape=dims,
                                   dtype='f',
                                   compression="gzip", compression_opts=9)
            h5group["LHC_noise_FB_bl"][0] = self.LHCNoiseFB.bl_meas

            if self.LHCNoiseFB.bl_meas_bbb is not None:

                h5group.create_dataset("LHC_noise_FB_bl_bbb",
                                       shape=(self.n_turns + 1,
                                              len(self.LHCNoiseFB.bl_meas_bbb)),
                                       dtype='f', compression="gzip",
                                       compression_opts=9)
                h5group["LHC_noise_FB_bl_bbb"][0,
                                               :] = self.LHCNoiseFB.bl_meas_bbb[:]

        # Close file
        self.close()

        # Initialise buffer for next turn
        self.init_buffer()

    def init_buffer(self):

        self.b_np_alive = np.zeros(self.buffer_time)
        self.b_mean_dt = np.zeros(self.buffer_time)
        self.b_mean_dE = np.zeros(self.buffer_time)
        self.b_sigma_dt = np.zeros(self.buffer_time)
        self.b_sigma_dE = np.zeros(self.buffer_time)
        self.b_epsn_rms = np.zeros(self.buffer_time)

        if self.fit_option:

            self.b_bl = np.zeros(self.buffer_time)

        if self.PL:

            self.b_PL_omegaRF = np.zeros(self.buffer_time)
            self.b_PL_phiRF = np.zeros(self.buffer_time)
            self.b_PL_bunch_phase = np.zeros(self.buffer_time)
            self.b_PL_phase_corr = np.zeros(self.buffer_time)
            self.b_PL_omegaRF_corr = np.zeros(self.buffer_time)
            self.b_SL_dphiRF = np.zeros(self.buffer_time)
            self.b_RL_drho = np.zeros(self.buffer_time)

        if self.LHCNoiseFB:

            self.b_LHCnoiseFB_factor = np.zeros(self.buffer_time)
            self.b_LHCnoiseFB_bl = np.zeros(self.buffer_time)
            if self.LHCNoiseFB.bl_meas_bbb is not None:
                self.b_LHCnoiseFB_bl_bbb = np.zeros((self.buffer_time,
                                                     len(self.LHCNoiseFB.bl_meas_bbb)))

    def write_buffer(self):

        i = self.i_turn % self.buffer_time

        self.b_np_alive[i] = self.beam.n_macroparticles_alive
        self.b_mean_dt[i] = self.beam.mean_dt
        self.b_mean_dE[i] = self.beam.mean_dE
        self.b_sigma_dt[i] = self.beam.sigma_dt
        self.b_sigma_dE[i] = self.beam.sigma_dE
        self.b_epsn_rms[i] = self.beam.epsn_rms_l

        if self.fit_option:

            self.b_bl[i] = self.profile.bunchLength

        if self.PL:

            self.b_PL_omegaRF[i] = self.rf_params.omega_rf[0, self.i_turn]
            self.b_PL_phiRF[i] = self.rf_params.phi_rf[0, self.i_turn]
            self.b_PL_bunch_phase[i] = self.PL.phi_beam
            self.b_PL_phase_corr[i] = self.PL.dphi
            self.b_PL_omegaRF_corr[i] = self.PL.domega_rf
            self.b_SL_dphiRF[i] = self.rf_params.dphi_rf[0]
            self.b_RL_drho[i] = self.PL.drho

        if self.LHCNoiseFB:

            self.b_LHCnoiseFB_factor[i] = self.LHCNoiseFB.x
            self.b_LHCnoiseFB_bl[i] = self.LHCNoiseFB.bl_meas
            if self.LHCNoiseFB.bl_meas_bbb is not None:
                self.b_LHCnoiseFB_bl_bbb[i, :] = self.LHCNoiseFB.bl_meas_bbb[:]

    def write_data(self, h5group, dims):

        i1 = self.i_turn - self.buffer_time
        i2 = self.i_turn

        h5group.require_dataset("n_macroparticles_alive", shape=dims,
                                dtype='f')
        h5group["n_macroparticles_alive"][i1:i2] = self.b_np_alive[:]

        h5group.require_dataset("mean_dt", shape=dims, dtype='f')
        h5group["mean_dt"][i1:i2] = self.b_mean_dt[:]

        h5group.require_dataset("mean_dE", shape=dims, dtype='f')
        h5group["mean_dE"][i1:i2] = self.b_mean_dE[:]

        h5group.require_dataset("sigma_dt", shape=dims, dtype='f')
        h5group["sigma_dt"][i1:i2] = self.b_sigma_dt[:]

        h5group.require_dataset("sigma_dE", shape=dims, dtype='f')
        h5group["sigma_dE"][i1:i2] = self.b_sigma_dE[:]

        h5group.require_dataset("epsn_rms_l", shape=dims, dtype='f')
        h5group["epsn_rms_l"][i1:i2] = self.b_epsn_rms[:]

        if self.fit_option:

            h5group.require_dataset("bunch_length", shape=dims,
                                    dtype='f')
            h5group["bunch_length"][i1:i2] = self.b_bl[:]

        if self.PL:

            h5group.require_dataset("PL_omegaRF", shape=dims,
                                    dtype=np.float64)
            h5group["PL_omegaRF"][i1:i2] = self.b_PL_omegaRF[:]

            h5group.require_dataset("PL_phiRF", shape=dims,
                                    dtype='f')
            h5group["PL_phiRF"][i1:i2] = self.b_PL_phiRF[:]

            h5group.require_dataset("PL_bunch_phase", shape=dims,
                                    dtype='f')
            h5group["PL_bunch_phase"][i1:i2] = self.b_PL_bunch_phase[:]

            h5group.require_dataset("PL_phase_corr", shape=dims,
                                    dtype='f')
            h5group["PL_phase_corr"][i1:i2] = self.b_PL_phase_corr[:]

            h5group.require_dataset("PL_omegaRF_corr", shape=dims,
                                    dtype='f')
            h5group["PL_omegaRF_corr"][i1:i2] = self.b_PL_omegaRF_corr[:]

            h5group.require_dataset("SL_dphiRF", shape=dims,
                                    dtype='f')
            h5group["SL_dphiRF"][i1:i2] = self.b_SL_dphiRF[:]

            h5group.require_dataset("RL_drho", shape=dims,
                                    dtype='f')
            h5group["RL_drho"][i1:i2] = self.b_RL_drho[:]

        if self.LHCNoiseFB:

            h5group.require_dataset("LHC_noise_FB_factor", shape=dims,
                                    dtype='f')
            h5group["LHC_noise_FB_factor"][i1:i2] = self.b_LHCnoiseFB_factor[:]

            h5group.require_dataset("LHC_noise_FB_bl", shape=dims,
                                    dtype='f')
            h5group["LHC_noise_FB_bl"][i1:i2] = self.b_LHCnoiseFB_bl[:]

            if self.LHCNoiseFB.bl_meas_bbb is not None:
                h5group.require_dataset("LHC_noise_FB_bl_bbb", shape=(self.n_turns + 1,
                                                                      len(self.LHCNoiseFB.bl_meas_bbb)),
                                        dtype='f')
                h5group["LHC_noise_FB_bl_bbb"][i1:i2,
                                               :] = self.b_LHCnoiseFB_bl_bbb[:, :]

    def open(self):
        self.h5file = hp.File(self.filename + '.h5', 'r+')
        self.h5file.require_group('Beam')

    def close(self):
        self.h5file.close()


class SlicesMonitor:

    ''' Class able to save the bunch profile, i.e. the histogram derived from
        the slicing.
    '''

    def __init__(self, filename, n_turns, profile):

        # create directory to save h5 file if needed
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

        self.h5file = hp.File(filename + '.h5', 'w')
        self.n_turns = n_turns
        self.i_turn = 0
        self.profile = profile
        self.h5file.create_group('Slices')

    def track(self, bunch):

        if not self.i_turn:
            self.create_data(self.h5file['Slices'], (self.profile.n_slices,
                                                     self.n_turns))
            self.write_data(self.profile, self.h5file['Slices'], self.i_turn)
        else:
            self.write_data(self.profile, self.h5file['Slices'], self.i_turn)

        self.i_turn += 1

    def create_data(self, h5group, dims):

        h5group.create_dataset("n_macroparticles", dims, compression="gzip",
                               compression_opts=9)

    def write_data(self, bunch, h5group, i_turn):

        h5group["n_macroparticles"][:, i_turn] = self.profile.n_macroparticles

    def close(self):
        self.h5file.close()


class MultiBunchMonitor:

    ''' Class able to save multi-bunch profile, i.e. the histogram derived from
        the slicing.
    '''

    def __init__(self, filename, n_turns, profile, rf, Nbunches, buffer_size=100):

        # create directory to save h5 file if needed
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

        self.h5file = hp.File(filename + '.h5', 'w')
        self.n_turns = n_turns
        self.i_turn = 0
        self.profile = profile
        self.rf = rf
        self.beam = self.profile.Beam
        self.h5file.create_group('default')
        self.h5group = self.h5file['default']
        self.Nbunches = Nbunches
        self.buffer_size = buffer_size
        self.last_save = 0

        self.create_data('profile', self.h5file['default'],
                         (self.n_turns, self.profile.n_slices), dtype='int32')
        self.b_profile = np.zeros(
            (self.buffer_size, self.profile.n_slices), dtype='int32')

        self.create_data(
            'turns', self.h5file['default'], (self.n_turns, ), dtype='int32')
        self.b_turns = np.zeros(self.buffer_size, dtype='int32')

        self.create_data('losses', self.h5file['default'],
                         (self.n_turns, ), dtype='int')
        self.b_losses = np.zeros(
            (self.buffer_size, ), dtype='int32')

        self.create_data('fwhm_bunch_position', self.h5file['default'],
                         (self.n_turns, self.Nbunches), dtype='float64')
        self.b_fwhm_bunch_position = np.zeros(
            (self.buffer_size, self.Nbunches), dtype=float)

        self.create_data('fwhm_bunch_length', self.h5file['default'],
                         (self.n_turns, self.Nbunches), dtype='float64')
        self.b_fwhm_bunch_length = np.zeros(
            (self.buffer_size, self.Nbunches), dtype=float)

        if self.Nbunches == 1:
            # All these can be calculated only when single bunch
            self.create_data(
                'mean_dE', self.h5file['default'], (
                    self.n_turns, self.Nbunches),
                dtype='float64')
            self.create_data(
                'dE_norm', self.h5file['default'], (
                    self.n_turns, self.Nbunches),
                dtype='float64')

            self.create_data(
                'mean_dt', self.h5file['default'], (
                    self.n_turns, self.Nbunches),
                dtype='float64')

            self.create_data(
                'dt_norm', self.h5file['default'], (
                    self.n_turns, self.Nbunches),
                dtype='float64')

            self.create_data(
                'std_dE', self.h5file['default'], (self.n_turns, self.Nbunches),
                dtype='float64')

            self.create_data(
                'std_dt', self.h5file['default'], (self.n_turns, self.Nbunches),
                dtype='float64')

            self.b_mean_dE = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)
            self.b_mean_dt = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)

            self.b_dE_norm = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)
            self.b_dt_norm = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)

            self.b_std_dE = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)
            self.b_std_dt = np.zeros(
                (self.buffer_size, self.Nbunches), dtype=float)

    def __del__(self):
        if self.i_turn > self.last_save:
            self.write_data()
        # self.h5file.close()

    def write_buffer(self, turn):

        # Nppb = int(self.profile.Beam.n_macroparticles // self.Nbunches)
        # mean_dE = np.zeros(self.Nbunches, dtype=float)
        # mean_dt = np.zeros(self.Nbunches, dtype=float)
        # std_dE = np.zeros(self.Nbunches, dtype=float)
        # std_dt = np.zeros(self.Nbunches, dtype=float)
        # for i in range(self.Nbunches):
        #     mean_dE[i] = np.mean(self.profile.Beam.dE[i*Nppb:(i+1)*Nppb])
        #     mean_dt[i] = np.mean(self.profile.Beam.dt[i*Nppb:(i+1)*Nppb])
        #     std_dE[i] = np.std(self.profile.Beam.dE[i*Nppb:(i+1)*Nppb])
        #     std_dt[i] = np.std(self.profile.Beam.dt[i*Nppb:(i+1)*Nppb])

        idx = self.i_turn % self.buffer_size

        self.b_turns[idx] = turn
        self.b_profile[idx] = self.profile.n_macroparticles.astype(np.int32)
        self.b_losses[idx] = self.beam.losses
        self.b_fwhm_bunch_position[idx] = self.profile.bunchPosition
        self.b_fwhm_bunch_length[idx] = self.profile.bunchLength

        if self.Nbunches == 1:
            self.b_mean_dE[idx] = self.beam.mean_dE
            self.b_mean_dt[idx] = self.beam.mean_dt
            self.b_std_dE[idx] = self.beam.sigma_dE
            self.b_std_dt[idx] = self.beam.sigma_dt
            self.b_dE_norm[idx] = self.rf.voltage[0, turn]

            if turn == 0:
                self.b_dt_norm[idx] = self.rf.t_rev[0] * self.rf.eta_0[0] * \
                    self.rf.voltage[0, 0] / \
                    (self.rf.beta[0]**2 * self.rf.energy[0])
            else:
                self.b_dt_norm[idx] = self.rf.t_rev[turn] * self.rf.eta_0[turn] * \
                    self.rf.voltage[0, turn - 1] / \
                    (self.rf.beta[turn]**2 * self.rf.energy[turn])

    def write_data(self):
        i1_h5 = self.last_save
        i2_h5 = self.i_turn
        i1_b = 0
        i2_b = self.i_turn - self.last_save
        # print("i1_h5, i2_h5:{}-{}".format(i1_h5, i2_h5))

        self.last_save = self.i_turn

        self.h5group['turns'][i1_h5:i2_h5] = self.b_turns[i1_b:i2_b]
        self.h5group['profile'][i1_h5:i2_h5] = self.b_profile[i1_b:i2_b]
        self.h5group['losses'][i1_h5:i2_h5] = self.b_losses[i1_b:i2_b]
        self.h5group['fwhm_bunch_position'][i1_h5:i2_h5] = self.b_fwhm_bunch_position[i1_b:i2_b]
        self.h5group['fwhm_bunch_length'][i1_h5:i2_h5] = self.b_fwhm_bunch_length[i1_b:i2_b]

        if self.Nbunches == 1:
            self.h5group['mean_dE'][i1_h5:i2_h5] = self.b_mean_dE[i1_b:i2_b]
            self.h5group['dE_norm'][i1_h5:i2_h5] = self.b_dE_norm[i1_b:i2_b]
            self.h5group['dt_norm'][i1_h5:i2_h5] = self.b_dt_norm[i1_b:i2_b]
            self.h5group['mean_dt'][i1_h5:i2_h5] = self.b_mean_dt[i1_b:i2_b]
            self.h5group['std_dE'][i1_h5:i2_h5] = self.b_std_dE[i1_b:i2_b]
            self.h5group['std_dt'][i1_h5:i2_h5] = self.b_std_dt[i1_b:i2_b]

    def track(self, turn):

        self.write_buffer(turn)
        self.i_turn += 1

        if (self.i_turn > 0) and (self.i_turn % self.buffer_size == 0):
            self.write_data()

    def create_data(self, name, h5group, dims, dtype):

        h5group.create_dataset(name, dims, compression='gzip',
                               compression_opts=4, dtype=dtype, shuffle=True)

    def close(self):
        if self.i_turn > self.last_save:
            self.write_data()
        self.h5file.close()




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Numerical diffusion model based on Ivanov (1992). Stationary single-harmonic
RF bucket is considered.**

:Authors: **Helga Timko**
'''

from __future__ import division

from builtins import range

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as int
from pylab import cm
from scipy.special import ellipk

from .action import x2, action_from_phase_amplitude


def phase_noise_diffusion(Ring, RFStation, spectrum, distribution,
                          distributionBins, Ngrids=200, M=1,
                          iterations=100000, figdir=None):
    '''
    Calculate diffusion in action space according to a given double-sided phase
    noise spectrum, on a uniform grid in oscillation amplitude.
    The spectrum is defined on the grid points (Ngrids + 1 points) for all M.
    The particle distribution in action is defined on the grids (Ngrids points). 
    Returns the diffused action distribution.
    Optional: define number of side-bands (M) to be taken into account, default
    is M = 1; N.B. this will only give impair modes for phase noise. 
    Optional: number of iterations to track.
    Optional: save figures into directory 'figdir'.
    '''

    # Input check
    N = Ngrids
    if spectrum.shape != (M, N + 1):
        # NoiseDiffusionError
        raise RuntimeError("In phase_noise_diffusion(): spectrum has to have shape (M, Ngrids+1)!")
    if len(distribution) != N:
        # NoiseDiffusionError
        raise RuntimeError("In phase_noise_diffusion(): distribution has to be an array of Ngrids elements!")

    # Some constants
    T0 = Ring.t_rev[0]
    omega_s0 = RFStation.omega_s0[0]
    h = RFStation.harmonic[0, 0]
    Jsep = 8. * omega_s0 / (np.pi * h**2)  # Action at the separatrix

    # Settings for plots
    plt.rc('axes', labelsize=16, labelweight='normal')
    plt.rc('lines', linewidth=1.5, markersize=6)
    plt.rc('font', family='sans-serif')
    plt.rc('legend', fontsize=12)

    # Construct action grid
    phimax = np.linspace(0., np.pi, N + 1, endpoint=True)
    xx = x2(phimax)
    J = action_from_phase_amplitude(xx)
    dJ = J[1:] - J[:-1]  # Differential on grid
    Jav = 0.5 * (J[1:] + J[:-1])  # Average on grid

    # Interpolate distribution
    distributionInterp = np.interp(Jav, distributionBins, distribution)

    # Normalise distribution
    distributionInterp /= int.simps(distributionInterp, Jav)

    # Construct weighting function
    Wm = np.zeros((M, N + 1))
    for k in range(0, M):
        m = 2 * k + 1
        Wm[k][:] = (np.pi * m / ellipk(xx))**4 / \
                   (4. * np.cosh(0.5 * np.pi * m * ellipk(1 - xx) / ellipk(xx))**2)

    # Diffusion coefficient for stationary bucket, according to Ivanov
    # Twice the sum over positive frequencies for double-sided spectrum
    D = (omega_s0 / h)**4 * np.sum(Wm * spectrum, axis=0)
    Dav = 0.5 * (D[1:] + D[:-1])  # Average on grid

    ax = plt.axes([0.15, 0.1, 0.8, 0.8])
    ax.plot(J, D)
    ax.set_xlabel(r"Relative action (J/J$_{\mathrm{sep}}$)")
    ax.set_ylabel(r"Diffusion coefficient [rad$^2$/s$^3$]")
    if figdir:
        plt.savefig(figdir + "D_vs_J.png")
        plt.clf()
    else:
        plt.show()

    ax = plt.axes([0.15, 0.1, 0.8, 0.8])
    for k in range(0, M):
        ax.plot(J, Wm[k], color=cm.get_cmap('jet')(k / M),
                label="m=%d mode" % (2 * k + 1))
    ax.set_xlabel(r"Relative action (J/J$_{\mathrm{sep}}$)")
    ax.set_ylabel(r"Weight function W$_m$ [1]")
    plt.legend(loc=0)
    if figdir:
        plt.savefig(figdir + "Wm_vs_J.png")
        plt.clf()
    else:
        plt.show()

    # Discretised diffusion equation ------------------------------------------
    A = np.zeros((N, N), float)
    for i in range(1, N - 1):
        A[i, i - 1] = dJ[i - 1]
        A[i, i] = 2. * (dJ[i - 1] + dJ[i])
        A[i, i + 1] = dJ[i]
    A[0, 0] = 2. * dJ[0]
    A[0, 1] = dJ[0]
    A[N - 1, N - 2] = dJ[N - 2]
    A[N - 1, N - 1] = 2. * (dJ[N - 2] + dJ[N - 1])
    A *= Jsep / 6.

    B = np.zeros((N, N), float)
    for i in range(1, N - 1):
        B[i, i - 1] = -Dav[i - 1] / dJ[i - 1]
        B[i, i] = Dav[i - 1] / dJ[i - 1] + Dav[i] / dJ[i]
        B[i, i + 1] = -Dav[i] / dJ[i]
    B[0, 0] = Dav[0] / dJ[0]
    B[0, 1] = -Dav[0] / dJ[0]
    B[N - 1, N - 2] = -Dav[N - 2] / dJ[N - 2]
    B[N - 1, N - 1] = Dav[N - 2] / dJ[N - 2] + Dav[N - 1] / dJ[N - 1]
    B /= Jsep

    # Time evolution ----------------------------------------------------------
    M1 = A - T0 * B / 2.
    M2 = A + T0 * B / 2.
    M1 = np.matrix(M1)
    M2 = np.matrix(M2)
    F = np.matrix(distributionInterp)
    Mtot = np.dot(M2.I, M1)
    Fold = F.T

    for i in range(0, iterations):

        Fnew = np.dot(Mtot, Fold)
        Fold = Fnew

    # Back to array for post-processing
    Fnew = Fnew.T
    F = np.array(F, order=0)
    Fnew = np.array(Fnew, order=0)

    # Plot --------------------------------------------------------------------
    # Distributions along action J
    norm_J_i = int.simps(F[0], Jav)
    norm_J_f = int.simps(Fnew[0], Jav)
    J_av_i = int.simps(F[0] * Jav, Jav) / norm_J_i
    J_av_f = int.simps(Fnew[0] * Jav, Jav) / norm_J_f

    # Conversion in SHORT-BUNCH APPROXIMATION!!!
    sigma_phi_i = np.sqrt(8. / np.pi * J_av_i)
    sigma_phi_f = np.sqrt(8. / np.pi * J_av_f)
    tau_i = sigma_phi_i * 2 * T0 / h / np.pi * 1.e9
    tau_f = sigma_phi_f * 2 * T0 / h / np.pi * 1.e9

    ax = plt.axes([0.12, 0.1, 0.78, 0.8])
    ax.plot(Jav, F[0], "b", label="Initial distribution")
    ax.plot(Jav, Fnew[0], "r", label="Final distribution")
    ax.set_xlabel(r"Relative action (J/J$_{\mathrm{sep}}$)")
    ax.set_ylabel("Particle distribution [1]")
    plt.legend(loc=1)
    ax2 = plt.twinx(ax)
    for k in range(0, M):
        ax2.plot(J, spectrum[k], color=cm.get_cmap('jet')(k / M), alpha=0.5)
        ax2.fill_between(J, 0, spectrum[k], color=cm.get_cmap('jet')(k / M),
                         alpha=0.2)
    ax2.set_ylabel(r"Double-sided spectral density [rad$^2$/Hz]")
    ax2.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    plt.figtext(0.6, 0.7, r'$\int_0^{J_{sep}}{F_i(J)}dJ=$ %.3f' % norm_J_i,
                fontsize=14, ha='left', va='center')
    plt.figtext(0.6, 0.625, r'$\int_0^{J_{sep}}{F_f(J)}dJ=$ %.3f' % norm_J_f,
                fontsize=14, ha='left', va='center')
    plt.figtext(0.6, 0.55, r'$<J>_i=$ %.4e s' % J_av_i, fontsize=14, ha='left',
                va='center')
    plt.figtext(0.6, 0.5, r'$<J>_f=$ %.4e s' % J_av_f, fontsize=14, ha='left',
                va='center')
    plt.figtext(0.4, 0.4, "Converting in short-bunch approximation...",
                fontsize=12, ha='left', va='center')
    plt.figtext(0.6, 0.35, r'$\sigma_{\varphi}^{(i)}=$ %.4f rad' % sigma_phi_i,
                fontsize=12, ha='left', va='center')
    plt.figtext(0.6, 0.3, r'$\sigma_{\varphi}^{(f)}=$ %.4f rad' % sigma_phi_f,
                fontsize=12, ha='left', va='center')
    plt.figtext(0.6, 0.25, r'$\tau_{4\sigma}^{(i)}=$ %.4f ns' % tau_i, fontsize=12,
                ha='left', va='center')
    plt.figtext(0.6, 0.2, r'$\tau_{4\sigma}^{(f)}=$ %.4f ns' % tau_f, fontsize=12,
                ha='left', va='center')
    if figdir:
        plt.savefig(figdir + "F_vs_J.png")
        plt.clf()
    else:
        plt.show()

    return Jav, distributionInterp, Fnew[0]



# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''

@author: Danilo Quartullo, Konstantinos Iliakis
'''

# MAKE SURE YOU HAVE GCC 4.8.1 OR LATER VERSIONS ON YOUR SYSTEM LINKED TO YOUR
# SYSTEM PATH.IF YOU ARE ON CERN-LXPLUS YOU CAN TYPE FROM CONSOLE
# source /afs/cern.ch/sw/lcg/contrib/gcc/4.8.1/x86_64-slc6/setup.sh
# TO GET GCC 4.8.1 64 BIT. IN GENERAL IT IS ADVISED TO USE PYTHON 64 BIT PLUS
# GCC 64 BIT.

from __future__ import print_function

import argparse
import ctypes
import os
import subprocess
import sys


def main():
    '''Compiles the blond C++ and/or CUDA library.
    '''
    path = os.path.realpath(__file__)
    basepath = os.sep.join(path.split(os.sep)[:-1])

    parser = argparse.ArgumentParser(
        description='Script used to compile the C++ (and CUDA) libraries needed by BLonD.',
        epilog='All arguments can be controlled with the environment variable BLOND_COMPILE_OPTS. E.g.: BLOND_COMPILE_OPTS=\'-p,--flags=-O0 -g\''
    )

    parser.add_argument('-p', '--parallel',
                        default=False, action='store_true',
                        help='Produce Multi-threaded code. Use the environment'
                        ' variable OMP_NUM_THREADS=xx to control the number of'
                        ' threads that will be used.'
                        ' Default: Serial code')

    parser.add_argument('-b', '--boost', type=str, nargs='?', const='',
                        help='Use boost library to speedup synchrotron radiation'
                        ' routines. If the installation path of boost differs'
                        ' from the default, you have to pass it as an argument.'
                        ' Default: Boost will not be used')

    parser.add_argument('-c', '--compiler', type=str, default='g++',
                        help='C++ compiler that will be used to compile the'
                        ' source files. Default: g++')

    parser.add_argument('--with-fftw', action='store_true',
                        help='Use the FFTs from FFTW3.')

    parser.add_argument('--with-fftw-threads', action='store_true',
                        help='Use the multi-threaded FFTs from FFTW3.')

    parser.add_argument('--with-fftw-omp', action='store_true',
                        help='Use the OMP FFTs from FFTW3.')

    parser.add_argument('--with-fftw-lib', type=str,
                        help='Path to the FFTW3 library (.so, .dll).')

    parser.add_argument('--with-fftw-header', type=str,
                        help='Path to the FFTW3 header files.')

    parser.add_argument('--flags', type=str, default='',
                        help='Additional compile flags.')

    parser.add_argument('--libs', type=str, default='',
                        help='Any extra libraries needed to compile')

    parser.add_argument('-libname', '--libname', type=str, default=os.path.join(basepath, 'cpp_routines/libblond'),
                        help='The C++ library name, without the file extension.')

    parser.add_argument('-optimize', '--optimize', action='store_true',
                        help='Auto optimize the compiled library.')

    parser.add_argument('-no-cpp', '--no-cpp', action='store_true',
                        help='Do not compile the C++ library.')

    parser.add_argument('-gpu', '--gpu', nargs='?', const='discover', default=None,
                        help='Compile the GPU kernels too.'
                        'Default: Only compile the C++ library.')

    parser.add_argument('-cuda-libname', '--cuda-libname', type=str, default=os.path.join(basepath, 'gpu/kernels'),
                        help='The CUDA library name, without the file extension.')

    # Additional libs needed to compile the blond library
    libs = []

    # EXAMPLE FLAGS: -Ofast -std=c++11 -fopt-info-vec -march=native
    #                -mfma4 -fopenmp -ftree-vectorizer-verbose=1 '-ffast-math'
    cflags = ['-O3', '-std=c++11', '-shared']
    # Some additional warning reporting related flags
    cflags += ['-Wall', '-Wno-unknown-pragmas']

    float_flags = ['-DUSEFLOAT']

    nvcc_flags = ['--cubin', '-O3', '--use_fast_math', '-maxrregcount', '32']

    cpp_files = [
        os.path.join(basepath, 'cpp_routines/kick.cpp'),
        os.path.join(basepath, 'cpp_routines/drift.cpp'),
        os.path.join(basepath, 'cpp_routines/linear_interp_kick.cpp'),
        os.path.join(basepath, 'cpp_routines/histogram.cpp'),
        os.path.join(basepath, 'cpp_routines/music_track.cpp'),
        os.path.join(basepath, 'cpp_routines/blondmath.cpp'),
        os.path.join(basepath, 'cpp_routines/fast_resonator.cpp'),
        os.path.join(basepath, 'cpp_routines/beam_phase.cpp'),
        os.path.join(basepath, 'cpp_routines/fft.cpp'),
        os.path.join(basepath, 'cpp_routines/openmp.cpp'),
        os.path.join(basepath, 'toolbox/tomoscope.cpp'),
        os.path.join(
            basepath, 'synchrotron_radiation/synchrotron_radiation.cpp'),
        os.path.join(basepath, 'beam/sparse_histogram.cpp'),
    ]

    cuda_files = [
        os.path.join(basepath, 'gpu/kernels.cu'),
    ]

    nvcc = 'nvcc'

    # Get nvcc from CUDA_PATH
    cuda_path = os.getenv('CUDA_PATH', default='')
    if cuda_path != '':
        nvcc = cuda_path + '/bin/nvcc'

    # Parse command line options
    args = vars(parser.parse_args())

    # Parse environment variable (BLOND_COMPILE_OPTS) options
    # if 'BLOND_COMPILE_OPTS' in os.environ:
    #     env_args_lst = os.environ['BLOND_COMPILE_OPTS'].split(',')
    #     env_args = vars(parser.parse_args(env_args_lst))
    #     args.update(env_args)

    if not args['no_cpp']:
        compile_cpp_library(args, cflags, float_flags, libs, cpp_files)

    if args['gpu']:
        compile_cuda_library(args, nvcc_flags, float_flags, cuda_files, nvcc)


def compile_cpp_library(args, cflags, float_flags, libs, cpp_files):
    # Check if we need to compile with FFTW
    with_fftw = args['with_fftw'] or args['with_fftw_threads'] or args['with_fftw_omp'] or \
        (args['with_fftw_lib'] is not None) or (args['with_fftw_header'] is not None)
    
    # Get boost path
    boost_path = None
    if args['boost'] is not None:
        if args['boost']:
            boost_path = os.path.abspath(args['boost'])
        else:
            boost_path = ''
        cflags += ['-I', boost_path, '-DBOOST']
    compiler = args['compiler']

    if args['libs']:
        libs = args['libs'].split()

    if args['parallel']:
        cflags += ['-fopenmp', '-DPARALLEL', '-D_GLIBCXX_PARALLEL']

    if args['flags']:
        cflags += args['flags'].split()

    if with_fftw:
        cflags += ['-DUSEFFTW3']
        if args['with_fftw_lib'] is not None:
            libs += ['-L', args['with_fftw_lib']]
        if args['with_fftw_header'] is not None:
            cflags += ['-I', args['with_fftw_header']]
        if 'win' in sys.platform:
            libs += ['-lfftw3-3']
        else:
            libs += ['-lfftw3', '-lfftw3f']
            if args['with_fftw_omp']:
                cflags += ['-DFFTW3PARALLEL']
                libs += ['-lfftw3_omp', '-lfftw3f_omp']
            elif args['with_fftw_threads']:
                cflags += ['-DFFTW3PARALLEL']
                libs += ['-lfftw3_threads', '-lfftw3f_threads']

    if 'posix' in os.name:
        cflags += ['-fPIC']
        if args['optimize']:
            if '-ffast-math' not in cflags:
                cflags += ['-ffast-math']
            # Check compiler defined directives
            # This is compatible with python3.6 - python 3.9
            # The universal_newlines argument transforms output to text (from binary)
            ret = subprocess.run(
                [compiler + ' -march=native -dM -E - < /dev/null | egrep "SSE|AVX|FMA"'],
                shell=True, stdout=subprocess.PIPE, universal_newlines=True,
                check=False)

            # If we have an error
            if ret.returncode != 0:
                print('Compiler auto-optimization did not work. Error: ', ret.stdout)
            else:
                # Format the output list
                stdout = ret.stdout.replace('#define ', '').replace(
                    '__ 1', '').replace('__', '').split('\n')
                # Add the appropriate vectorization flag (not use avx512)
                if 'AVX2' in stdout:
                    cflags += ['-mavx2']
                elif 'AVX' in stdout:
                    cflags += ['-mavx']
                elif 'SSE4_2' in stdout or 'SSE4_1' in stdout:
                    cflags += ['-msse4']
                elif 'SSE3' in stdout:
                    cflags += ['-msse3']
                else:
                    cflags += ['-msse']

                # Add FMA if supported
                if 'FMA' in stdout:
                    cflags += ['-mfma']

        root, ext = os.path.splitext(args['libname'])
        if not ext:
            ext = '.so'
        libname_single = os.path.abspath(root + '_single' + ext)
        libname_double = os.path.abspath(root + '_double' + ext)

    elif 'win' in sys.platform:
        root, ext = os.path.splitext(args['libname'])
        if not ext:
            ext = '.dll'

        libname_single = os.path.abspath(root + '_single' + ext)
        libname_double = os.path.abspath(root + '_double' + ext)

        if hasattr(os, 'add_dll_directory'):
            directory, _ = os.path.split(libname_double)
            os.add_dll_directory(directory)

    else:
        print(
            'YOU ARE NOT USING A WINDOWS OR LINUX OPERATING SYSTEM. ABORTING...')
        sys.exit(-1)


    # Report the compilation options
    print('Enable Multi-threaded code: ', args['parallel'])
    print('Use boost: ', args['boost'] is not None)
    if args['boost'] is not None:
        print('Boost installation path: ', boost_path)
    print('Link with FFTW3: ', with_fftw)
    if with_fftw:
        print('Parallel FFTW3:',
              args['with_fftw_threads'] or args['with_fftw_omp'])
    if args['with_fftw_lib'] or args['with_fftw_header']:
        print('FFTW3 Library path: ', args['with_fftw_lib'])
        print('FFTW3 Headers path: ', args['with_fftw_header'])
    print('C++ Compiler: ', compiler)
    compiler_version = subprocess.run([compiler, '--version'],
                                      capture_output=True,
                                      check=False).stdout.decode().split('\n')[0]
    print('Compiler version: ', compiler_version)

    print('Compiler flags: ', ' '.join(cflags))
    print('Extra libraries: ', ' '.join(libs))

    command = [compiler] + cflags + float_flags + \
        cpp_files + libs + ['-o', libname_single]
    print('\nCompiling the single-precision (32-bit) C++ library')
    ret = run_compile(command, libname_single)
    if ret != 0:
        print('There was a compilation error.')
    else:
        # Verify that the libraries have been compiled
        try:
            if ('win' in sys.platform) and hasattr(os, 'add_dll_directory'):
                _ = ctypes.CDLL(libname_single, winmode=0)
            else:
                _ = ctypes.CDLL(libname_single)
            print('Compiled successfully.')
        except Exception as exception:
            print('Compilation failed.')
            print(exception)

    command = [compiler] + cflags + cpp_files + libs + ['-o', libname_double]
    print('\nCompiling the double-precision (64-bit) C++ library')
    ret = run_compile(command, libname_double)
    if ret != 0:
        print('There was a compilation error.')
    else:
        # Verify that the libraries have been compiled
        try:
            if ('win' in sys.platform) and hasattr(os, 'add_dll_directory'):
                _ = ctypes.CDLL(libname_double, winmode=0)
            else:
                _ = ctypes.CDLL(libname_double)
            print('Compiled successfully.')
        except Exception as exception:
            print('Compilation failed.')
            print(exception)


def compile_cuda_library(args, nvccflags, float_flags, cuda_files, nvcc):
    # Compile the GPU library
    # print('\n' + ''.join(['='] * 80))
    print('\nCompiling the CUDA library')
    import cupy as cp

    if args['gpu'] == 'discover':
        print('Discovering the device compute capability..')

        dev = cp.cuda.Device(0)
        dev_name = cp.cuda.runtime.getDeviceProperties(dev)['name']
        comp_capability = dev.compute_capability
        print(f'Device name {dev_name}')
    elif args['gpu'] is not None:
        comp_capability = args['gpu']

    print(
        f'Compiling the CUDA library for compute capability {comp_capability}.')

    # Add the -arch required argument
    nvccflags += ['-arch', f'sm_{comp_capability}']
    
    # Get the CuPy header files location
    path = cp.__file__.split('/')[:-1]  # remove __init__.py from path
    path.extend(['_core', 'include'])
    cupyloc = os.path.join('/'.join(path))

    
    print('CUDA Compiler: ', nvcc)
    compiler_version = subprocess.run([nvcc, '--version'],
                                      capture_output=True,
                                      check=False).stdout.decode().split('\n')[0]
    print('Compiler version: ', compiler_version)
    print('Compiler flags: ', ' '.join(nvccflags))
    print('CuPy location: ', cupyloc)

    
    libname_double = args['cuda_libname'] + f'_sm_{comp_capability}_double.cubin'
    libname_single = args['cuda_libname'] + f'_sm_{comp_capability}_single.cubin'

    command = [nvcc] + nvccflags + \
        ['-o', libname_single, '-I' + cupyloc] + float_flags + cuda_files
    
    print('\nCompiling the single-precision (32-bit) CUDA library')
    ret = run_compile(command, libname_single)
    if ret != 0:
        print('There was a compilation error.')
    else:
        print('Compiled successfully.')

    command = [nvcc] + nvccflags + \
        ['-o', libname_double, '-I' + cupyloc] + cuda_files
    print('\nCompiling the double-precision (64-bit) CUDA library')
    ret = run_compile(command, libname_double)
    if ret != 0:
        print('There was a compilation error.')
    else:
        print('Compiled successfully.')


def run_compile(command, libname):
    if os.path.exists(libname):
        os.remove(libname)
    print(' '.join(command))
    ret = subprocess.run(command, check=False)
    if ret.returncode != 0 or not os.path.isfile(libname):
        return -1
    else:
        return 0


if __name__ == "__main__":
    main()




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3), 
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities 
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to plot different bunch features**

:Authors: **Helga Timko**, **Danilo Quartullo**
'''

from __future__ import division
import matplotlib.pyplot as plt
import numpy as np



def plot_beam_profile(Profile, counter, style = '-', dirname = 'fig', show_plot = False):
    
    """
    Plot of longitudinal beam profile
    """
 
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(Profile.bin_centers, Profile.n_macroparticles, style)
    
    ax.set_xlabel(r"$\Delta t$ [s]")
    ax.set_ylabel('Beam profile [arb. units]')
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    
    plt.figtext(0.95, 0.95, '%d turns' %counter, fontsize=16, ha='right', 
                va='center') 
    
    # Save plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/beam_profile_' "%d" %counter + '.png'
        plt.savefig(fign)
    plt.clf()



def plot_beam_profile_derivative(Profile, counter, style='-', dirname='fig', show_plot = False,
                                 modes=['diff']):
    """
    Plot of the derivative of the longitudinal beam profile.
    Modes list should contain 1 or more of the elements below:
    1) 'filter1d', 2) 'gradient', 3) 'diff'
    """
    for mode in modes:
        x, derivative = Profile.beam_profile_derivative(mode)
        plt.plot(x, derivative, style, label=mode)
    plt.legend()
    if show_plot:
        plt.show()
    else:
        fign = dirname + '/beam_profile_derivative_' "%d" % counter + '.png'
        plt.savefig(fign)
    plt.clf()
    

def plot_beam_spectrum(Profile, counter, style = '-', dirname = 'fig', show_plot = False):
    
    """
    Plot of longitudinal beam profile
    """
 
    plt.figure(1, figsize=(8,6))
    ax = plt.axes()
    ax.plot(Profile.beam_spectrum_freq, np.absolute(Profile.beam_spectrum), style)
    
    ax.set_xlabel(r"Frequency [Hz]")
    ax.set_ylabel('Beam spectrum, \n absolute value [arb. units]')
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.set_xlim(0,5.e9)
    
    plt.figtext(0.95, 0.95, '%d turns' %counter, fontsize=16, ha='right', 
                va='center') 
    
    # Save plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/beam_spectrum_' "%d" %counter + '.png'
        plt.savefig(fign)
    plt.clf()
